from ._version import (__version__, __updated__)

__copyright__     = "Copyright (C) 2013-2015 European Commission (JRC)"
__license__       = "EUPL 1.1+"

__title__ = "pandalone"
__summary__ = "Collection of utilities for working with hierarchical data with relocatable paths."
__uri__ = "https://github.com/pandalone/pandalone"



#! python
#-*- coding: utf-8 -*-
#
# Copyright 2015 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

from __future__ import division, unicode_literals

import doctest
import json
from pandalone.mappings import (
    Pmod, pmods_from_tuples, Pstep, _join_paths, _append_step)
import pandalone.mappings
import re
import sre_constants
import sys
import unittest

from numpy import testing as npt

import functools as ft
import pandas as pd


def pmod2regexstrs(pmod):
    if pmod._regxs:
        return [r.pattern for r in list(pmod._regxs.keys())]


@unittest.skipIf(sys.version_info < (3, 4), "Doctests are made for py >= 3.3")
class Doctest(unittest.TestCase):

    def test_doctests(self):
        failure_count, test_count = doctest.testmod(
            pandalone.mappings,
            optionflags=doctest.NORMALIZE_WHITESPACE)  # | doctest.ELLIPSIS)
        self.assertGreater(test_count, 0, (failure_count, test_count))
        self.assertEquals(failure_count, 0, (failure_count, test_count))


class TestJoinPaths(unittest.TestCase):

    def _check_append_step(self, all_steps, exp_steps):
        steps = all_steps[:-1]
        step = all_steps[-1]
        msg = "{} + {} != {}".format(steps, step, exp_steps)
        self.assertEqual(_append_step(steps, step), exp_steps, msg)

    def _masscheck_append_step(self, cases):
        for all_steps, exp_steps in cases:
            self._check_append_step(all_steps, exp_steps)

    def test_append_step_simple(self):
        cases = [
            [('a',), ('a',)],
            [('a', 'b', '..'), ('a',)],
            [('a', 'b', '.'), ('a', 'b')],
        ]
        self._masscheck_append_step(cases)

    def test_append_step_absolute(self):
        cases = [
            [('a', 'b', ''), ('',)],
            [('', ''), ('',)],
            [('',), ('',)],
        ]
        self._masscheck_append_step(cases)

    def test_append_step_dotdot(self):
        cases = [
            [('', '..'), ('',)],
            [('', '.'), ('',)],
            [('a', '..'), ()],
            [('..',), ('..',)],
            [('..', '..'), ('..', '..')],
            [('.',), ()],
        ]
        self._masscheck_append_step(cases)

    def test_join_paths_dotdot_on_empty(self):
        self.assertEqual(_join_paths('..',), '..')
        self.assertEqual(_join_paths('a', '..', '..'), '..')
        self.assertEqual(_join_paths('..'), '..')

    def test_join_paths_dotdot_on_rooted(self):
        self.assertEqual(_join_paths('', '..'), '',)

    def test_join_paths_root_end(self):
        self.assertEqual(_join_paths('', 'a', ''), '')
        self.assertEqual(_join_paths('', ''), '')
        self.assertEqual(_join_paths(''), '')


class TestPmod(unittest.TestCase):

    def setUp(self):
        self.assertPmod_class_attributes_not_modified()

    def tearDown(self):
        self.assertPmod_class_attributes_not_modified()

    def assertPmod_class_attributes_not_modified(self):
        self.assertEqual(
            Pmod.__init__.__defaults__, (None, {}, {}))  # @UndefinedVariable

    def test_merge_name(self):
        pm1 = Pmod(_alias='pm1')
        pm2 = Pmod(_alias='pm2')
        pm = pm1._merge(pm2)
        self.assertEqual(pm._alias, 'pm2')
        self.assertEqual(pm1._alias, 'pm1')
        self.assertEqual(pm2._alias, 'pm2')
        pm = pm2._merge(pm1)
        self.assertEqual(pm._alias, 'pm1')
        self.assertEqual(pm1._alias, 'pm1')
        self.assertEqual(pm2._alias, 'pm2')

        pm1 = Pmod()
        pm2 = Pmod(_alias='pm2')
        pm = pm1._merge(pm2)
        self.assertEqual(pm._alias, 'pm2')
        self.assertEqual(pm1._alias, None)
        self.assertEqual(pm2._alias, 'pm2')
        pm = pm2._merge(pm1)
        self.assertEqual(pm._alias, 'pm2')
        self.assertEqual(pm1._alias, None)
        self.assertEqual(pm2._alias, 'pm2')

    def test_merge_name_recurse(self):
        pm1 = Pmod(_alias='pm1', _steps={'a': Pmod(_alias='R1')})
        pm2 = Pmod(_alias='pm2', _steps={'a': Pmod(_alias='R2'),
                                         'b': Pmod(_alias='R22')})
        pm = pm1._merge(pm2)
        self.assertEqual(pm._steps['a']._alias, 'R2')
        self.assertEqual(pm1._steps['a']._alias, 'R1')
        self.assertEqual(pm2._steps['a']._alias, 'R2')
        self.assertEqual(len(pm1._steps), 1)
        self.assertEqual(len(pm2._steps), 2)
        pm = pm2._merge(pm1)
        self.assertEqual(pm._steps['a']._alias, 'R1')
        self.assertEqual(pm1._steps['a']._alias, 'R1')
        self.assertEqual(pm2._steps['a']._alias, 'R2')
        self.assertEqual(len(pm1._steps), 1)
        self.assertEqual(len(pm2._steps), 2)

    def test_merge_steps(self):
        pm1 = Pmod(_alias='pm1', _steps={'a': Pmod(_alias='A'),
                                         'c': Pmod(_alias='C')})
        pm2 = Pmod(_alias='pm2', _steps={'b': Pmod(_alias='B'),
                                         'a': Pmod(_alias='AA'),
                                         'd': Pmod(_alias='DD'),
                                         })
        pm = pm1._merge(pm2)
        self.assertEqual(sorted(pm._steps.keys()), list('abcd'))
        pm = pm2._merge(pm1)
        self.assertEqual(sorted(pm._steps.keys()), list('abcd'))
        self.assertEqual(len(pm1._steps), 2)
        self.assertEqual(len(pm2._steps), 3)

        pm1 = Pmod(_steps={'a': Pmod(_alias='A'),
                           'c': Pmod(_alias='C')})
        pm2 = Pmod(_alias='pm2', _steps={'b': Pmod(_alias='B'),
                                         'a': Pmod(_alias='AA'),
                                         'd': Pmod(_alias='DD'),
                                         })
        pm = pm1._merge(pm2)
        self.assertEqual(sorted(pm._steps.keys()), list('abcd'))
        self.assertEqual(pm._steps['a']._alias, 'AA')
        self.assertEqual(len(pm1._steps), 2)
        self.assertEqual(len(pm2._steps), 3)
        self.assertEqual(pm._regxs, {})
        self.assertEqual(pm1._regxs, {})
        self.assertEqual(pm2._regxs, {})

        pm = pm2._merge(pm1)
        self.assertEqual(sorted(pm._steps.keys()), list('abcd'))
        self.assertEqual(pm._steps['a']._alias, 'A')
        self.assertEqual(len(pm1._steps), 2)
        self.assertEqual(len(pm2._steps), 3)
        self.assertEqual(pm._regxs, {})
        self.assertEqual(pm1._regxs, {})
        self.assertEqual(pm2._regxs, {})

    def test_merge_regxs(self):
        pm1 = Pmod(_alias='pm1', _regxs=[
            ('e', Pmod(_alias='E')),
            ('a', Pmod(_alias='A')),
            ('c', Pmod(_alias='C'))])
        pm2 = Pmod(_alias='pm2', _regxs=[
            ('b', Pmod(_alias='B')),
            ('a', Pmod(_alias='AA')),
            ('d', Pmod(_alias='DD')),
        ])

        pm = pm1._merge(pm2)
        self.assertSequenceEqual(pmod2regexstrs(pm), list('ecbad'))
        self.assertEqual(pm._regxs[re.compile('a')]._alias, 'AA')
        self.assertSequenceEqual(pmod2regexstrs(pm1), list('eac'))
        self.assertSequenceEqual(pmod2regexstrs(pm2), list('bad'))
        self.assertEqual(pm._steps, {})
        self.assertEqual(pm1._steps, {})
        self.assertEqual(pm2._steps, {})

        pm = pm2._merge(pm1)
        self.assertSequenceEqual(pmod2regexstrs(pm), list('bdeac'))
        self.assertEqual(pm._regxs[re.compile('a')]._alias, 'A')
        self.assertSequenceEqual(pmod2regexstrs(pm1), list('eac'))
        self.assertSequenceEqual(pmod2regexstrs(pm2), list('bad'))
        self.assertEqual(pm._steps, {})
        self.assertEqual(pm1._steps, {})
        self.assertEqual(pm2._steps, {})

    def test_merge_all(self):
        def merge_all(pmods):
            return ft.reduce(Pmod._merge, pmods)

        pm1 = Pmod(_steps={'a': Pmod(_alias='A')})
        pm2 = Pmod(_alias='pm2',
                   _regxs=[
                       ('b', Pmod(_alias='BB')),
                       ('a', Pmod(_alias='AA'))
                   ])
        pm3 = Pmod(_alias='PM3',
                   _steps={'c': Pmod(_alias='CCC'),
                           'b': Pmod(_alias='BBB'), },
                   _regxs=[
                       ('b', Pmod(_alias='AAA')),
                       ('a', Pmod(_alias='BBB')),
                       ('c', Pmod(_alias='CCC')),
                   ])
        pm = merge_all([pm1, pm2, pm3])

        self.assertSetEqual(set(pm._steps.keys()), set(list('abc')))
        self.assertSetEqual(set(pm1._steps.keys()), set(list('a')))
        self.assertEqual(pm2._steps, {})
        self.assertSetEqual(set(pm3._steps.keys()), set(list('bc')))

        self.assertEqual(pmod2regexstrs(pm), list('bac'))
        self.assertEqual(pm1._regxs, {})
        self.assertEqual(pmod2regexstrs(pm2), list('ba'))
        self.assertEqual(pmod2regexstrs(pm3), list('bac'))

    def test_descend_None(self):
        pm = Pmod()
        self.assert_descend_stops(pm.descend('a'))

        pm = Pmod(_alias='a')
        self.assert_descend_stops(pm.descend('a'))

        pm = Pmod(_steps={'a': None})
        self.assert_descend_stops(pm.descend('a'))

    def test_alias_None(self):
        pm = Pmod()
        self.assertIsNone(pm.alias('a'))

        pm = Pmod(_alias='a')
        self.assertIsNone(pm.alias('a'))

        pm = Pmod(_steps={'a': None})
        self.assertIsNone(pm.alias('a'))

    def _build_pmod_manually(self):
        return Pmod(
            _steps={'a': Pmod(_alias='A')},
            _regxs=[('a(\w*)', Pmod('AWord')),
                    ('a(\d*)', Pmod(_alias='A_\\1')), ])

    def assert_descend_stops(self, cpmod, msg=None):
        self.assertEqual(cpmod, (None, None), msg)

    def test_descend_BAD(self):
        pm = self._build_pmod_manually()
        self.assert_descend_stops(pm.descend('BAD'))
        self.assert_descend_stops(pm.descend('a-'))
        self.assert_descend_stops(pm.descend('a$'))

    def test_alias_BAD(self):
        pm = self._build_pmod_manually()
        self.assertIsNone(pm.alias('BAD'))
        self.assertIsNone(pm.alias('a-'))
        self.assertIsNone(pm.alias('a$'))

    def test_descend(self):
        pm = Pmod(
            _steps={'a':
                    Pmod(_alias='A', _steps={1: 11})},
            _regxs=[
                ('a\w*', Pmod(_alias='AWord', _steps={2: 22})),
                ('a\d*', Pmod(_alias='ADigit', _steps={3: 33})),
            ])

        # All children and regexps match.
        self.assertDictEqual(pm.descend('a')[0]._steps, {1: 11, 2: 22, 3: 33})

        # Only 'a\w*' matches.
        self.assertDictEqual(pm.descend('aa')[0]._steps, {2: 22})

        # Both regexps matches.
        self.assertDictEqual(pm.descend('a1')[0]._steps, {2: 22, 3: 33})

    def test_alias(self):
        pm = self._build_pmod_manually()

        self.assertEqual(pm.alias('a'), 'A')
        self.assertEqual(pm.alias('abc'), 'AWord')
        self.assertEqual(pm.alias('a12'), 'A_12')
        self.assertIsNone(pm.alias('BAD'))

    def test_descend_group_zero_(self):
        pm = Pmod(_regxs=[('abc', Pmod(_alias='A\\g<0>'))])
        self.assertEqual(pm.descend('abc')[1], 'Aabc')

        pm = Pmod(_regxs=[('a(.*)', Pmod(_alias='A\\g<0>'))])
        self.assertEqual(pm.descend('abc')[1], 'Aabc')

    def test_alias_group_zero_(self):
        pm = Pmod(_regxs=[('abc', Pmod(_alias='A\\g<0>'))])
        self.assertEqual(pm.alias('abc'), 'Aabc')

        pm = Pmod(_regxs=[('a(.*)', Pmod(_alias='A\\g<0>'))])
        self.assertEqual(pm.alias('abc'), 'Aabc')

    def test_alias_invalid_group_ref_(self):
        pm = Pmod(_regxs=[
            ('a', Pmod(_alias='A\\1')),
            ('b(.*)', Pmod(_alias='B\\2')),
        ])

        with self.assertRaises(sre_constants.error):
            pm.alias('a')

        with self.assertRaises(sre_constants.error):
            pm.alias('bach')

    def test_map_path_rootUnmapped_slashPath(self):
        pm = Pmod()
        self.assertEqual(pm.map_path('/'), '')

    def test_map_path_rootUnmapped_emptyPath(self):
        pm = Pmod()
        self.assertEqual(pm.map_path(''), '')

    def test_map_path_rootMapped(self):

        pm = pmods_from_tuples([
            ('', 'New/Root'),
            ('/a/b', 'B'),
        ])
        self.assertEqual(pm.map_path('/a/b'), 'New/Root/a/B')

    def _build_simple_pmod_from_tuples(self):
        return pmods_from_tuples([('/a', 'A'),
                                  ('/~a(\w*)', 'AWord'),
                                  ('/~a(\d*)', 'A_\\1'),
                                  ])

    def test_map_path_rootMapped_relative(self):
        pm = pmods_from_tuples([('', 'root')])
        self.assertEqual(pm.map_path(''), 'root')
        self.assertEqual(pm.map_path('/a'), 'root/a')
        # self.assertEqual(pm.map_path('/'), 'root')  ## TODO: FIX map2slash

        pm = pmods_from_tuples([('', 'root'), ('/', 'root2')])
        self.assertEqual(pm.map_path(''), 'root')
        self.assertEqual(pm.map_path('/a'), 'root/a')
        self.assertEqual(pm.map_path('/'), 'root/root2')

    def test_map_path_rootMapped_absolute(self):
        pm = pmods_from_tuples([('', '/root')])
        self.assertEqual(pm.map_path(''), '/root')
        self.assertEqual(pm.map_path('/a'), '/root/a')
        # self.assertEqual(pm.map_path('/'), '/root')  ## TODO: FIX map2slash

        pm = pmods_from_tuples([('', '/root'), ('/', 'root2')])
        self.assertEqual(pm.map_path(''), '/root')
        self.assertEqual(pm.map_path('/a'), '/root/a')
        self.assertEqual(pm.map_path('/'), '/root/root2')

    def test_map_path_missed(self):
        pm = self._build_simple_pmod_from_tuples()

        self.assertEqual(pm.map_path('/c'), '/c')
        self.assertEqual(pm.map_path('/c/a'), '/c/a')
        self.assertEqual(pm.map_path('/a$/a4'), '/a$/a4')

    def test_map_path_simple(self):
        pm = self._build_simple_pmod_from_tuples()

        self.assertEqual(pm.map_path('/a'), '/A')
        self.assertEqual(pm.map_path('/a/b'), '/A/b')

    def test_map_path_regex(self):
        pm = self._build_simple_pmod_from_tuples()

        self.assertEqual(pm.map_path('/a'), '/A')
        self.assertEqual(pm.map_path('/a_wo/b'), '/AWord/b')
        self.assertEqual(pm.map_path('/a12/b'), '/A_12/b')

    def test_map_path_regex_2_dot_dotdot_root_abs(self):
        pmods = pmods_from_tuples([('/~b.*/~c(.*)',  r'../C/\1')])
        self.assertEqual(pmods.map_path('/'), '')
        self.assertEqual(pmods.map_path('/a'), '/a')
        self.assertEqual(pmods.map_path('/big/child'), '/big/C/hild')
        self.assertEqual(pmods.map_path(''), '')

    def test_build_pmods_steps_no_regxs(self):
        pmods_tuples = [
            ('/a', 'A1/A2'),
            ('/a/b', 'B'),
        ]
        pmods = pmods_from_tuples(pmods_tuples)

        pmods = pmods._steps['']
        self.assertFalse(bool(pmods._regxs))
        self.assertSetEqual(set(pmods._steps), {'a'})
        self.assertEqual(pmods._steps['a']._alias, 'A1/A2')
        self.assertEqual(pmods._steps['a']._steps['b']._alias, 'B')

    def test_build_pmods_regex_no_steps(self):
        pmods_tuples = [
            ('/~a*', 'A1/A2'),
            ('/~a[1]?/~b?', 'B'),
        ]
        pmods = pmods_from_tuples(pmods_tuples)

        pmods = pmods._steps['']
        self.assertFalse(bool(pmods._steps))
        self.assertEqual(pmod2regexstrs(pmods), ['a*', 'a[1]?'])

    @unittest.skipIf(sys.version_info < (3, 4), "String repr differ in py-2.")
    def test_pmods_from_tuples_repr(self):
        pmods_tuples = [
            ('/a', 'A'),
            ('/~a\w*', 'A1/A2'),
            ('/~a\d*/~b?', 'D/D'),
        ]
        pmods = pmods_from_tuples(pmods_tuples)

        pmods = pmods._steps['']
        s = [r"pmod({'a': pmod('A')}, OrderedDict([(re.compile('a\\w*'), "
             r"pmod('A1/A2')), (re.compile('a\\d*'), "
             r"pmod(OrderedDict([(re.compile('b?'), pmod('D/D'))])))]))"
             ]
        self.assertEqual(str(pmods), "".join(s))

    def test_pmods_from_tuples_rootMapping(self):
        self.assertEqual(pmods_from_tuples([('', 'A/B')]),
                         Pmod(_steps={'': Pmod(_alias='A/B')}))
        self.assertEqual(pmods_from_tuples([('/', 'A/B')]),
                         Pmod(_steps={'':
                                      Pmod(_steps={'': Pmod(_alias='A/B')})}))

    def test_map_path_2_root(self):
        pmods = pmods_from_tuples([('/a', '')])
        self.assertEqual(pmods.map_path('/a'), '')
        self.assertEqual(pmods.map_path('/a/b'), '/b')
        self.assertEqual(pmods.map_path('/b'), '/b')

        pmods = pmods_from_tuples([('/a/b', '')])
        self.assertEqual(pmods.map_path('/a'), '/a')
        self.assertEqual(pmods.map_path('/a/b'), '')
        self.assertEqual(pmods.map_path('/a/b/c'), '/c')
        self.assertEqual(pmods.map_path('/c'), '/c')

    def test_map_path_2_abs(self):
        pmods = pmods_from_tuples([('/a', '/A')])
        self.assertEqual(pmods.map_path('/a'), '/A')
        self.assertEqual(pmods.map_path('/a/b'), '/A/b')
        self.assertEqual(pmods.map_path('/A/b'), '/A/b')

        pmods = pmods_from_tuples([('/a/b', '/A/AA')])
        self.assertEqual(pmods.map_path('/a'), '/a')
        self.assertEqual(pmods.map_path('/a/b'), '/A/AA')
        self.assertEqual(pmods.map_path('/a/b/c'), '/A/AA/c')
        self.assertEqual(pmods.map_path('/A/AA/c'), '/A/AA/c')

    def test_map_path_2_dot(self):
        pmods = pmods_from_tuples([('/a', '.')])
        self.assertEqual(pmods.map_path('/a'), '/a')
        self.assertEqual(pmods.map_path('/a/b'), '/a/b')
        self.assertEqual(pmods.map_path('/b'), '/b')

        pmods = pmods_from_tuples([('/a', './bb')])
        self.assertEqual(pmods.map_path('/a'), '/a/bb')
        self.assertEqual(pmods.map_path('/a/b'), '/a/bb/b')
        self.assertEqual(pmods.map_path('/b/c'), '/b/c')

    def test_map_path_2_dotdot(self):
        pmods = pmods_from_tuples([('/a', '..')])
        self.assertEqual(pmods.map_path('/a'), '')
        self.assertEqual(pmods.map_path('/a/b'), '/b')
        self.assertEqual(pmods.map_path('/b'), '/b')

        pmods = pmods_from_tuples([('/a/b', '..')])
        self.assertEqual(pmods.map_path('/a'), '/a')
        self.assertEqual(pmods.map_path('/a/b'), '/a')
        self.assertEqual(pmods.map_path('/a/b/c'), '/a/c')

        pmods = pmods_from_tuples([('/a/b', '../cc')])
        self.assertEqual(pmods.map_path('/a'), '/a')
        self.assertEqual(pmods.map_path('/a/b'), '/a/cc')
        self.assertEqual(pmods.map_path('/a/b/c'), '/a/cc/c')

    def test_map_path_slash_2_dot_dotdot_root_abs(self):
        pmods = pmods_from_tuples([('/', '.')])
        self.assertEqual(pmods.map_path('/'), '')
        self.assertEqual(pmods.map_path('/a'), '/a')
        self.assertEqual(pmods.map_path('/a/b'), '/a/b')
        self.assertEqual(pmods.map_path(''), '')

        pmods = pmods_from_tuples([('/', '..')])
        self.assertEqual(pmods.map_path('/'), '')
        self.assertEqual(pmods.map_path('/a'), '/a')
        self.assertEqual(pmods.map_path('/a/b'), '/a/b')
        self.assertEqual(pmods.map_path(''), '')

        pmods = pmods_from_tuples([('/', '')])
        self.assertEqual(pmods.map_path('/'), '')
        self.assertEqual(pmods.map_path('/a'), '/a')
        self.assertEqual(pmods.map_path('/a/b'), '/a/b')
        self.assertEqual(pmods.map_path(''), '')

        pmods = pmods_from_tuples([('/', '/')])
        self.assertEqual(pmods.map_path('/'), '')
        self.assertEqual(pmods.map_path('/a'), '/a')
        self.assertEqual(pmods.map_path('/a/b'), '/a/b')
        self.assertEqual(pmods.map_path(''), '')

        pmods = pmods_from_tuples([('/', '/A/B')])
        self.assertEqual(pmods.map_path('/'), '/A/B')
        self.assertEqual(pmods.map_path('/a'), '/a')
        self.assertEqual(pmods.map_path('/a/b'), '/a/b')
        self.assertEqual(pmods.map_path(''), '')

    def test_map_path_root_2_dot_dotdot_root_abs(self):
        pmods = pmods_from_tuples([('', '.')])
        self.assertEqual(pmods.map_path('/'), '')
        self.assertEqual(pmods.map_path('/a'), '/a')
        self.assertEqual(pmods.map_path('/a/b'), '/a/b')
        self.assertEqual(pmods.map_path(''), '')

        pmods = pmods_from_tuples([('', '..')])
        self.assertEqual(pmods.map_path('/'), '')
        self.assertEqual(pmods.map_path('/a'), '/a')
        self.assertEqual(pmods.map_path('/a/b'), '/a/b')
        self.assertEqual(pmods.map_path(''), '')

        pmods = pmods_from_tuples([('', '/')])
        self.assertEqual(pmods.map_path('/'), '')
        self.assertEqual(pmods.map_path('/a'), '/a')
        self.assertEqual(pmods.map_path('/a/b'), '/a/b')
        self.assertEqual(pmods.map_path(''), '')

        pmods = pmods_from_tuples([('', '/A/B')])
        #self.assertEqual(pmods.map_path('/'), '/A/B')
        self.assertEqual(pmods.map_path('/a'), '/A/B/a')
        self.assertEqual(pmods.map_path('/a/b'), '/A/B/a/b')
        self.assertEqual(pmods.map_path(''), '/A/B')

    def test_pmods_from_tuple_ignores_empties(self):
        # Ignored by :func:`pmods_from_tuples().
        pmods = pmods_from_tuples([('', '')])
        self.assertEqual(pmods, Pmod())
        pmods = pmods_from_tuples([('spam', None)])
        self.assertEqual(pmods, Pmod())
        pmods = pmods_from_tuples([(None, 'spam')])
        self.assertEqual(pmods, Pmod())


class TestPstep(unittest.TestCase):

    def PMODS(self, absolute=False):
        return pmods_from_tuples([
            ('',            '/root' if absolute else 'root'),
            ('/a',          'b'),
            ('/abc',        'BAR'),
            ('/abc/df',     'DEF'),
            ('/abc/df/a123', '234'),
            ('/some',       '/sub/path'),
        ])

    def test_equality(self):
        p = Pstep()
        self.assertEqual(str(p), '')
        self.assertEqual(p, Pstep(''))

        self.assertEqual(str(p.a), 'a')
        self.assertEqual(p.a, Pstep('a'))

        n = 'foo'
        p = Pstep(n)
        self.assertEqual(str(p), n)
        self.assertEquals(p, Pstep(n))
        self.assertEquals(str(p.foo), str(p))

        n = '/foo'
        p = Pstep(n)
        self.assertEqual(str(p), n)
        self.assertEquals(p, Pstep(n))

    def test_buildtree_valid_ops(self):
        p = Pstep()
        p.abc
        p.abc.ddd = 'def'
        p.abc.dfg = 'defg'
        p.n123
        p.pp = '321'
        p._some_hidden = '12'
        p.a.b.c.d = '123'
        p.a.b.c.d.e = '123'

        return p

    def test_constructor(self):
        pmod = pmods_from_tuples([('/a', 'A'), ('/a/c', 'C'), ('/b/c', 'C1')])
        p = pmod.step()
        self.assertEquals(p.a, 'A')
        self.assertEquals(p.a.c, 'C')
        self.assertEqual(p.b, 'b')
        self.assertEqual(p.b.c, 'C1')

        p.a = 'AA'
        self.assertEqual(p.a, 'AA')
        self.assertIn('AA', p._steps())
        self.assertIn('c', p.a._steps(keys=True))
        self.assertIn('C', p.a._steps())
        self.assertIn('c', p.b._steps(keys=True))
        self.assertIn('C1', p.b._steps())

    def test_buildtree_invalid_ops(self):

        p = Pstep()

        def f1(p):
            p.abc = 1

        def f2(p):
            p.a = 1

        def f3(p):
            p.a.b = 1

        def f4(p):
            p.a.c = 1

        for f in [f1, f2, f3, f4]:
            p = Pstep()
            with self.assertRaises(AssertionError, msg=f):
                f(p),

    def test_paths_empty1ststep(self):
        p = Pstep()
        self.assertListEqual(sorted(p._paths()), sorted(['']))

        p.a
        self.assertListEqual(sorted(p._paths()), sorted(['/a']))

        p = Pstep('')
        self.assertListEqual(sorted(p._paths()), sorted(['']))

        p = Pstep('')
        p.a
        self.assertListEqual(sorted(p._paths()), sorted(['/a']))

    def test_paths(self):
        p = Pstep('r')
        self.assertListEqual(sorted(p._paths()), sorted(['r']))

        p.a
        self.assertListEqual(sorted(p._paths()), sorted(['r/a']))

    def test_paths_multi_empty1ststep(self):
        p = Pstep()
        p.abc
        p.abc.aaa = 'def'
        p.abc.defg
        p.n123
        p.pp = '321'
        p._some_hidden = 12
        exp = [
            '/abc/def',
            '/abc/defg',
            '/321',
            '/n123',
        ]
        self.assertListEqual(sorted(p._paths()), sorted(exp))

        exp = [
            '/abc/(aaa-->def)',
            '/abc/defg',
            '/(pp-->321)',
            '/n123',
        ]
        self.assertListEqual(sorted(p._paths(with_orig=True)), sorted(exp))

    def test_paths_multi_nonemptyroot(self):
        p = Pstep('r')
        p.abc
        p.abc.aaa = 'def'
        p.abc.defg
        p.n123
        p.pp = '321'
        p._some_hidden = 12
        exp = [
            'r/abc/def',
            'r/abc/defg',
            'r/321',
            'r/n123',
        ]
        self.assertListEqual(sorted(p._paths()), sorted(exp))

        exp = [
            'r/abc/(aaa-->def)',
            'r/abc/defg',
            'r/(pp-->321)',
            'r/n123',
        ]

    ### DOT ###

    def test_paths_dot_atroot_emptyroot(self):
        p = Pstep()
        p.pp = '.'
        self.assertListEqual(sorted(p._paths()), sorted(['']))

    def test_paths_dot_emptyroot(self):
        p = Pstep()
        p.a.b = '.'
        self.assertListEqual(sorted(p._paths()), sorted(['/a']))

    def test_paths_dot_atroot(self):
        p = Pstep('r')
        p.pp = '.'
        self.assertListEqual(sorted(p._paths()), sorted(['r']))
        p.pp = '.'
        self.assertListEqual(sorted(p._paths()), sorted(['r']))

    def test_paths_dot(self):
        p = Pstep('r')
        p.a.b = '.'
        self.assertListEqual(sorted(p._paths()), sorted(['r/a']))

    def test_paths_dot_multi_emptyroot(self):
        p = Pstep('')
        p.pp = '.'
        p.pp.a.b
        p.c.pp = '.'
        p.c.pp.d
        self.assertListEqual(sorted(p._paths()), sorted(['/a/b', '/c/d']))

    def test_paths_dot_multi(self):
        p = Pstep('r')
        p.pp = '.'
        p.pp.a.b
        p.c.cc = '.'
        p.c.cc.d
        self.assertListEqual(sorted(p._paths()), sorted(['r/a/b', 'r/c/d']))

    ### DOTDOT ###
    # NOT TO BE USED DIRECTLY
    # BC HIDS PATHS,
    # but testing for mapping.

    def test_paths_dotdot_atroot_emptyroot(self):
        p = Pstep()
        p.pp = '..'
        p.pp.a
        self.assertListEqual(sorted(p._paths()), sorted(['/a']))

        p = Pstep()
        p.pp = '..'
        p.pp.a
        p.pp.b
        self.assertListEqual(sorted(p._paths()), sorted(['/a', '/b']))

    def test_paths_dotdot_emptyroot(self):
        p = Pstep()
        p.a.aa = '..'
        p.a.aa.b
        self.assertListEqual(sorted(p._paths()), sorted(['/b']))

        p = Pstep()
        p.a.aa = '..'
        p.a.aa.b
        p.a.c
        self.assertListEqual(sorted(p._paths()), sorted(['/a/c', '/b']))

    def test_paths_dotdot_atroot(self):
        p = Pstep('r')
        p.pp = '..'
        p.pp.a
        self.assertListEqual(sorted(p._paths()), sorted(['a']))

    def test_paths_dotdot(self):
        p = Pstep('r')
        p.a.aa = '..'
        p.a.aa.b
        self.assertListEqual(sorted(p._paths()), sorted(['r/b']))

    def test_paths_dotdot_multi_emptyroot(self):
        p = Pstep('')
        p.pp = '..'
        p.pp.a.b
        p.c.cc = '..'
        p.c.cc.d
        self.assertListEqual(sorted(p._paths()), sorted(['/a/b', '/d']))

    def test_paths_dotdot_multi(self):
        p = Pstep('r')
        p.pp = '..'
        p.pp.a.b
        p.c.cc = '..'
        p.c.cc.d
        self.assertListEqual(sorted(p._paths()), sorted(['a/b', 'r/d']))

    ### PMODS ###

    def test_pmods_miss(self):
        p = pmods_from_tuples([('/MISS', 'BOO')]).step()
        self.assertEquals(p, '', (p, p._paths()))
        p = Pstep('foo', pmods_from_tuples([('/MISS', 'BOO')]))
        self.assertEquals(p, 'foo', (p, p._paths()))

    def test_pmods_emptyroot_rootunmapped(self):
        p = pmods_from_tuples([('/a', 'bar')]).step()
        p.a
        self.assertEquals(p, '', (p, p._paths()))
        self.assertEquals(p.a, 'bar', (p, p._paths()))
        self.assertEquals(sorted(p._paths()), ['/bar'])

        self.assertEquals(p.a.b, 'b')
        self.assertEquals(sorted(p._paths()), ['/bar/b'])

        self.assertEquals(p.c, 'c')
        self.assertEquals(sorted(p._paths()), ['/bar/b', '/c'])

        self.assertEquals(p.a.f, 'f')
        self.assertEquals(sorted(p._paths()), ['/bar/b', '/bar/f', '/c'])

    def test_pmods_nonemptyroot_rootunmapped(self):
        p = pmods_from_tuples([('root/a', 'bar')]).step('root')
        p.a
        self.assertEquals(p, 'root', (p, p._paths()))
        self.assertEquals(p.a, 'bar', (p, p._paths()))
        self.assertEquals(sorted(p._paths()), ['root/bar'])

        self.assertEquals(p.a.b, 'b')
        self.assertEquals(sorted(p._paths()), ['root/bar/b'])

        self.assertEquals(p.c, 'c')
        self.assertEquals(sorted(p._paths()), ['root/bar/b', 'root/c'])

        self.assertEquals(p.a.f, 'f')
        self.assertEquals(sorted(p._paths()),
                          ['root/bar/b', 'root/bar/f', 'root/c'])

    def test_pmods_emptyroot_rootmapped(self):
        p = pmods_from_tuples([('', 'root'), ('/a', 'bar')]).step()
        p.a
        self.assertEquals(p, 'root')
        self.assertEquals(p.a, 'bar')
        self.assertEquals(sorted(p._paths()), ['root/bar'])

        self.assertEquals(p.a.b, 'b')
        self.assertEquals(sorted(p._paths()), ['root/bar/b'])

        self.assertEquals(p.c, 'c')
        self.assertEquals(sorted(p._paths()), ['root/bar/b', 'root/c'])

        self.assertEquals(p.a.f, 'f')
        self.assertEquals(
            sorted(p._paths()), ['root/bar/b', 'root/bar/f', 'root/c'])

    def test_pmods_nonemptyroot_rootmapped(self):
        p = pmods_from_tuples(
            [('root', 'ROOT'), ('root/a', 'bar')]).step('root')
        p.a
        self.assertEquals(p, 'ROOT')
        self.assertEquals(p.a, 'bar')
        self.assertEquals(sorted(p._paths()), ['ROOT/bar'])

        self.assertEquals(p.a.b, 'b')
        self.assertEquals(sorted(p._paths()), ['ROOT/bar/b'])

        self.assertEquals(p.c, 'c')
        self.assertEquals(sorted(p._paths()), ['ROOT/bar/b', 'ROOT/c'])

        self.assertEquals(p.a.f, 'f')
        self.assertEquals(
            sorted(p._paths()), ['ROOT/bar/b', 'ROOT/bar/f', 'ROOT/c'])

    def _build_psteps(self, root='', pmods=None):
        p = Pstep(root, pmods)
        p.a.b.c
        p.a.b.d
        p.a.c
        p.abc.df
        p.n123
        p.a.n123
        p.cc.a123.a123
        p.cc.a123.abc
        p.b321
        # p[''] ##NOTE: Will always return to root.

        return p

    def _assert_pstep_pmods_with_map_paths(self, root, pmods):
        ps1 = self._build_psteps(root, pmods=pmods)
        ps2 = self._build_psteps(root)

        self.assertEqual(sorted(ps1._paths()),
                         sorted(pmods.map_paths(ps2._paths())))

    def test_pstep_vs_pmods_maproot(self):
        root = 'R'
        pmods = pmods_from_tuples([(root, 'AA'), ])
        self._assert_pstep_pmods_with_map_paths(root, pmods)

    def test_pstep_vs_pmods_mapslash(self):
        pmods = pmods_from_tuples([('/', 'AA'), ])
        self._assert_pstep_pmods_with_map_paths('P', pmods)

    def test_pstep_vs_pmods_mapstep(self):

        pmods = pmods_from_tuples([('/a', 'AA'), ])
        self._assert_pstep_pmods_with_map_paths('P', pmods)

    def test_pstep_vs_pmods_maproot_emptyroot(self):
        pmods = pmods_from_tuples([('', 'AA'), ])
        self._assert_pstep_pmods_with_map_paths('', pmods)

    def test_pstep_vs_pmods_mapslash_empty1ststep(self):
        pmods = pmods_from_tuples([('/', 'AA'), ])
        self._assert_pstep_pmods_with_map_paths('', pmods)

    def test_pstep_vs_pmods_mapstep_empty1ststep(self):
        pmods = pmods_from_tuples([('/a', 'AA'), ])
        self._assert_pstep_pmods_with_map_paths('', pmods)

    def test_paths_pmods_mass(self):
        p = self.PMODS().step()
        p.a
        p.abc.df.a123
        exp = ['root/b', 'root/BAR/DEF/234']
        self.assertListEqual(sorted(p._paths()), sorted(exp), (p, p._paths()))

        p = self.PMODS(absolute=True).step()
        p.a
        p.abc.df
        p.abc.df.a123
        exp = ['/root/b', '/root/BAR/DEF/234']
        self.assertListEqual(sorted(p._paths()), sorted(exp), (p, p._paths()))

    def test_paths_pmods_orig_mass(self):
        p = self.PMODS().step()
        p.a
        p.abc.df.a123
        exp = [
            '(-->root)/(a-->b)',
            '(-->root)/(abc-->BAR)/(df-->DEF)/(a123-->234)'
        ]
        self.assertListEqual(
            sorted(p._paths(with_orig=True)), sorted(exp), (p, p._paths()))

        p = self.PMODS(absolute=True).step()
        p.a
        p.abc.df.a123
        exp = [
            '(-->/root)/(a-->b)',
            '(-->/root)/(abc-->BAR)/(df-->DEF)/(a123-->234)'
        ]
        self.assertListEqual(
            sorted(p._paths(with_orig=True)), sorted(exp), (p, p._paths()))

    def test_pmods_lock_not_applying(self):
        p = self.PMODS().step('not dot')
        p.nota
        p = self.PMODS(absolute=True).step('not dot')
        p.nota

        pmods = self.PMODS()
        pmods._alias = None
        p = pmods.step()
        pmods = self.PMODS(absolute=True)
        pmods._alias = None
        p = pmods.step()

    def test_pmods_lock_CAN_RELOCATE_unmapped(self):
        p = Pstep()
        p._locked = Pstep.CAN_RELOCATE
        p2 = p._lock
        self.assertIs(p2, p)
        p2 = p._fix
        self.assertIs(p2, p)

        p2 = p.abc._lock
        self.assertIs(p2, p.abc)
        p2 = p.abc._fix
        self.assertIs(p2, p.abc)

    def test_pmods_lock_CAN_RELOCATE_mapped(self):
        pmods = self.PMODS()
        p = pmods.step()
        p._locked = Pstep.CAN_RELOCATE

        p2 = p.unmapped._lock
        self.assertIs(p2, p.unmapped)
        p2 = p.unmapped._fix
        self.assertIs(p2, p.unmapped)

    def test_pmods_lock_CAN_RENAME_root(self):
        pmods = self.PMODS(absolute=True)
        p = pmods.step()
        with self.assertRaises(ValueError, msg=p._paths()):
            p._locked = Pstep.CAN_RENAME
        with self.assertRaises(ValueError, msg=p._paths()):
            p._lock

    def test_pmods_lock_CAN_RENAME_rest(self):
        p = self.PMODS().step()
        p2 = p.a._fix  # `a` just renamed `b`.
        self.assertIs(p2, p.a)
        with self.assertRaises(ValueError, msg=p._paths()):
            p.some._locked = Pstep.CAN_RENAME
        with self.assertRaises(ValueError, msg=p._paths()):
            p.some._lock

    def test_pmods_lock_LOCKED_root(self):
        p = self.PMODS().step()
        with self.assertRaises(ValueError, msg=p._paths()):
            p._locked = Pstep.LOCKED
        with self.assertRaises(ValueError, msg=p._paths()):
            p._lock

    def test_pmods_lock_LOCKED_rest(self):
        pmods = self.PMODS()
        p = pmods.step()
        with self.assertRaises(ValueError, msg=p._paths()):
            p.abc._locked = Pstep.LOCKED
        with self.assertRaises(ValueError, msg=p._paths()):
            p.abc._lock

        with self.assertRaises(ValueError, msg=p._paths()):
            p.some._locked = Pstep.LOCKED
        with self.assertRaises(ValueError, msg=p._paths()):
            p.some._lock

    def test_map_from_pstep(self):
        p = self.test_buildtree_valid_ops()
        pairs = p._derrive_map_tuples()
        #print('\n'.join(str(p) for p in pairs))

        exp = [
            ('', ''),
            ('/a', 'a'),
            ('/a/b', 'b'),
            ('/a/b/c', 'c'),
            ('/a/b/c/d', '123'),
            ('/a/b/c/d/e', '123'),
            ('/abc', 'abc'),
            ('/abc/ddd', 'def'),
            ('/abc/dfg', 'defg'),
            ('/n123', 'n123'),
            ('/pp', '321'),
        ]
        self.assertEqual(pairs, exp)

    def test_assign(self):
        p1 = Pstep('root')

        def f1():
            p1.a = 1

        p1.a = 'a'
        p1.a = p1
        p1.a = Pstep()  # TODO: Make step assigning work?

        self.assertRaises(AssertionError, f1)

    def test_schema(self):
        json.dumps(Pstep())
        json.dumps({Pstep(): 1})

    def test_json(self):
        p = Pstep()
        p._schema.allOf = {}
        p._schema.type = 'list'
        p.a._schema.kws = {'minimum': 1}

    def test_json_list(self):
        json.dumps([Pstep(), Pstep('a')])

    def test_tags_paths(self):
        p = Pstep()
        self.assertEqual(p._tags, ())
        self.assertEqual(p._paths(tag='BAD'), [])

        p._tag('inp')
        self.assertSequenceEqual(sorted(p._tags), ['inp'])
        self.assertEqual(p._paths(tag='BAD'), [])
#         self.assertEqual(p._paths(tag='inp'), [''])

        self.assertEqual(p.a._tags, ())

        p.a._tag('t')
        self.assertSequenceEqual(sorted(p._paths(tag='t')), ['/a'])
        self.assertEqual(p._paths(tag='BAD'), [])
        p.b
        self.assertSequenceEqual(sorted(p._paths(tag='t')), ['/a'])
        p.a.aa
        self.assertSequenceEqual(sorted(p._paths(tag='t')), ['/a'])
        p.b.c
        self.assertSequenceEqual(sorted(p._paths(tag='t')), ['/a'])
        p.b._tag('t')
        self.assertSequenceEqual(sorted(p._paths(tag='t')), ['/a', '/b'])
        p.b.c._tag('t')
        self.assertSequenceEqual(
            sorted(p._paths(tag='t')), ['/a', '/b', '/b/c'])
        p.a.aa._tag('nt')
        self.assertSequenceEqual(
            sorted(p._paths(tag='t')), ['/a', '/b', '/b/c'])
        p.a.aa._tag('t')
        self.assertSequenceEqual(
            sorted(p._paths(tag='t')), ['/a', '/a/aa', '/b', '/b/c'])

    def test_tags_delete(self):
        p = Pstep()
        p.a._tag('u').aa._tag('t').aaa
        p.b.c._tag('u')
        self.assertSequenceEqual(sorted(p._paths()), ['/a/aa/aaa', '/b/c'])
        self.assertSequenceEqual(sorted(p._paths(tag='t')), ['/a/aa'])
        self.assertSequenceEqual(sorted(p._paths(tag='u')), ['/a', '/b/c'])
        self.assertSequenceEqual(sorted(p._paths(tag='BAD')), [])

        p.a._tag_remove('u').aa.aaa
        self.assertSequenceEqual(sorted(p._paths()), ['/a/aa/aaa', '/b/c'])
        self.assertSequenceEqual(sorted(p._paths(tag='t')), ['/a/aa'])
        self.assertSequenceEqual(sorted(p._paths(tag='u')), ['/b/c'])

        p.a.aa._tag_remove('t').aaa
        p.a.aa._tag_remove('u').aaa  # Useless
        p.b.c._tag_remove('u')
        self.assertSequenceEqual(sorted(p._paths()), ['/a/aa/aaa', '/b/c'])
        self.assertSequenceEqual(sorted(p._paths(tag='t')), [])
        self.assertSequenceEqual(sorted(p._paths(tag='u')), [])

        p.a.aa.aaa._tag_remove('u')  # Useless
        p.a.aa.aaa._tag_remove('t')
        self.assertSequenceEqual(sorted(p._paths()), ['/a/aa/aaa', '/b/c'])
        self.assertSequenceEqual(sorted(p._paths(tag='t')), [])
        self.assertSequenceEqual(sorted(p._paths(tag='u')), [])

    def test_as_pandas_index(self):
        df = pd.DataFrame([1, 3], columns=['a'])
        s = Pstep('a')
        npt.assert_array_equal(df[s], [1, 3])

    def test_from_pandas(self):
        cdf = pd.DataFrame([
            ('fc',      'CO2 [g/km]'),
            ('jobno',   'Job [-]'),
        ], columns=['colkey', 'colname'])
        cdf['colkey'] = '/' + cdf['colkey']
        c = pmods_from_tuples(cdf[['colkey', 'colname']].values).step()
        self.assertEqual(c.fc, 'CO2 [g/km]', c._csteps)
        self.assertEqual(c.jobno, 'Job [-]', c._csteps)


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# Copyright (c) 2013-2015 Julian Berman, Kostis Anagnostopoulos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import division, print_function, unicode_literals

from collections import deque
from contextlib import contextmanager
import json
from jsonschema.validators import (
    RefResolutionError, UnknownType, RefResolver
)
import unittest

from jsonschema import FormatChecker, ValidationError

import numpy as np
import numpy.testing as npt
from pandalone.pandata import PandelVisitor
import pandas as pd


try:
    from unittest import mock
except ImportError:
    import mock


def validate(instance, schema, *args, **kws):
    return PandelVisitor(schema, *args, **kws).validate(instance)


def wrap_as_sequences(seq):
    """
    Accepts a sequence and yields it like (list, tuple, ndarray),
    """


def wrap_in_pandas(dict_or_list, wrap_as_df=False):
    """
    Accepts a mapping and yields it like (dict, Series),
    """
    if isinstance(dict_or_list, dict):
        for op in (dict, pd.Series):
            yield op(dict_or_list)
        if wrap_as_df:
            yield pd.DataFrame(dict_or_list)
    else:
        for op in (list, tuple, np.array):
            yield op(dict_or_list)


class TestIterErrors(unittest.TestCase):

    def iter_errors(self, instance, schema, *args, **kwds):
        self.validator = PandelVisitor(schema)
        return self.validator.iter_errors(instance, *args, **kwds)

    #@unittest.skip("For Draft3Validator only!")
    def test_iter_errors(self):
        data = [1, 2]
        for instance in wrap_in_pandas(data):
            schema = {
                "$schema": "http://json-schema.org/draft-03/schema#",
                u"disallow": u"array",
                u"enum": [["a", "b", "c"], ["d", "e", "f"]],
                u"minItems": 3
            }

            got = (e for e in self.iter_errors(instance, schema))
            expected = [
                "disallow",
                "minItems",
                "enum",
            ]
            self.assertListEqual(
                sorted(e.validator for e in got), sorted(expected))

    def test_iter_errors_multiple_failures_one_validator(self):
        tree1 = {"foo": 2, "bar": [1], "baz": 15, "quux": "spam"}
        tree2 = {"foo": 2, "bar": np.array([1]), "baz": 15, "quux": "spam"}
        for data in (tree1, tree2,):
            for instance in wrap_in_pandas(data):
                schema = {
                    u"properties": {
                        "foo": {u"type": "string"},
                        "bar": {u"minItems": 2},
                        "baz": {u"maximum": 10, u"enum": [2, 4, 6, 8]},
                    }
                }

                errors = list(self.iter_errors(instance, schema))
                self.assertEqual(len(errors), 4, errors)


class TestValidationErrorMessages(unittest.TestCase):

    def message_for(self, instance, schema, *args, **kwargs):
        with self.assertRaises(ValidationError) as e:
            validate(instance, schema, *args, **kwargs)
        return e.exception.message

    def test_single_type_failure(self):
        message = self.message_for(instance=1, schema={u"type": u"string"})
        self.assertEqual(message, "1 is not of type %r" % u"string")

    def test_single_type_list_failure(self):
        message = self.message_for(instance=1, schema={u"type": [u"string"]})
        self.assertEqual(message, "1 is not of type %r" % u"string")

    def test_multiple_type_failure(self):
        types = u"string", u"object"
        message = self.message_for(instance=1, schema={u"type": list(types)})
        self.assertEqual(message, "1 is not of type %r, %r" % types)

    #@unittest.skip("For Draft3Validator only!")
    def test_object_without_title_type_failure(self):
        atype = {
            u"type": [{u"minimum": 3}],
        }
        schema = {
            "$schema": "http://json-schema.org/draft-03/schema#",
            u"type": [atype]
        }
        message = self.message_for(instance=1, schema=schema)
        self.assertEqual(message, "1 is not of type %r" % (atype,))

    #@unittest.skip("For Draft3Validator only!")
    def test_object_with_name_type_failure(self):
        name = "Foo"
        schema = {
            u"type": [{u"name": name, u"minimum": 3}],
            "$schema": "http://json-schema.org/draft-03/schema#"
        }
        message = self.message_for(instance=1, schema=schema)
        self.assertEqual(message, "1 is not of type %r" % (name,))

    def test_minimum(self):
        message = self.message_for(instance=1, schema={"minimum": 2})
        self.assertEqual(message, "1 is less than the minimum of 2")

    def test_maximum(self):
        message = self.message_for(instance=1, schema={"maximum": 0})
        self.assertEqual(message, "1 is greater than the maximum of 0")

    def test_dependencies_failure_has_single_element_not_list(self):
        depend, on = "bar", "foo"
        schema = {
            u"dependencies": {depend: on},
            "$schema": "http://json-schema.org/draft-03/schema#"
        }
        data = {"bar": 2}
        for instance in wrap_in_pandas(data):
            message = self.message_for(instance, schema)
            self.assertEqual(
                message, "%r is a dependency of %r" % (on, depend))

    def test_additionalItems_single_failure(self):
        data = [2]
        for instance in wrap_in_pandas(data):
            message = self.message_for(
                instance, {
                    u"items": [], u"additionalItems": False,
                    "$schema": "http://json-schema.org/draft-03/schema#"
                },
            )
        self.assertIn("(2 was unexpected)", message)

    def test_additionalItems_multiple_failures(self):
        data = [1, 2, 3]
        for instance in wrap_in_pandas(data):
            message = self.message_for(
                instance, {
                    "$schema": "http://json-schema.org/draft-03/schema#",
                    u"items": [], u"additionalItems": False
                }
            )
            self.assertIn("(1, 2, 3 were unexpected)", message)

    def test_additionalProperties_single_failure(self):
        additional = "foo"
        schema = {u"additionalProperties": False}
        data = {additional: 2}
        for instance in wrap_in_pandas(data):
            message = self.message_for(instance, schema)
            self.assertIn("(%r was unexpected)" % (additional,), message)

    def test_additionalProperties_multiple_failures(self):
        schema = {u"additionalProperties": False}
        data = ["foo", "bar"]
        for instance in wrap_in_pandas(data):
            message = self.message_for(dict.fromkeys(instance), schema)

            self.assertIn(repr("foo"), message)
            self.assertIn(repr("bar"), message)
            self.assertIn("were unexpected)", message)

    def test_invalid_format_default_message(self):
        checker = FormatChecker(formats=())
        check_fn = mock.Mock(return_value=False)
        checker.checks(u"thing")(check_fn)

        schema = {u"format": u"thing"}
        message = self.message_for("bla", schema, format_checker=checker)

        self.assertIn(repr("bla"), message)
        self.assertIn(repr("thing"), message)
        self.assertIn("is not a", message)


class TestValidationErrorDetails(unittest.TestCase):
    # TODO: These really need unit tests for each individual rule, rather
    #       than just these higher level tests.

    def test_anyOf(self):
        instance = 5
        schema = {
            "anyOf": [
                {"minimum": 20},
                {"type": "string"}
            ]
        }

        validator = PandelVisitor(schema)
        errors = list(validator.iter_errors(instance))
        self.assertEqual(len(errors), 1)
        e = errors[0]

        self.assertEqual(e.validator, "anyOf")
        self.assertEqual(e.validator_value, schema["anyOf"])
        self.assertEqual(e.instance, instance)
        self.assertEqual(e.schema, schema)
        self.assertIsNone(e.parent)

        self.assertEqual(e.path, deque([]))
        self.assertEqual(e.relative_path, deque([]))
        self.assertEqual(e.absolute_path, deque([]))

        self.assertEqual(e.schema_path, deque(["anyOf"]))
        self.assertEqual(e.relative_schema_path, deque(["anyOf"]))
        self.assertEqual(e.absolute_schema_path, deque(["anyOf"]))

        self.assertEqual(len(e.context), 2)

        e1, e2 = sorted_errors(e.context)

        self.assertEqual(e1.validator, "minimum")
        self.assertEqual(e1.validator_value, schema["anyOf"][0]["minimum"])
        self.assertEqual(e1.instance, instance)
        self.assertEqual(e1.schema, schema["anyOf"][0])
        self.assertIs(e1.parent, e)

        self.assertEqual(e1.path, deque([]))
        self.assertEqual(e1.absolute_path, deque([]))
        self.assertEqual(e1.relative_path, deque([]))

        self.assertEqual(e1.schema_path, deque([0, "minimum"]))
        self.assertEqual(e1.relative_schema_path, deque([0, "minimum"]))
        self.assertEqual(
            e1.absolute_schema_path, deque(["anyOf", 0, "minimum"]),
        )

        self.assertFalse(e1.context)

        self.assertEqual(e2.validator, "type")
        self.assertEqual(e2.validator_value, schema["anyOf"][1]["type"])
        self.assertEqual(e2.instance, instance)
        self.assertEqual(e2.schema, schema["anyOf"][1])
        self.assertIs(e2.parent, e)

        self.assertEqual(e2.path, deque([]))
        self.assertEqual(e2.relative_path, deque([]))
        self.assertEqual(e2.absolute_path, deque([]))

        self.assertEqual(e2.schema_path, deque([1, "type"]))
        self.assertEqual(e2.relative_schema_path, deque([1, "type"]))
        self.assertEqual(e2.absolute_schema_path, deque(["anyOf", 1, "type"]))

        self.assertEqual(len(e2.context), 0)

    #@unittest.skip("For Draft3Validator only!")
    def test_type(self):
        instance = {"foo": 1}
        schema = {
            "$schema": "http://json-schema.org/draft-03/schema#",
            "type": [
                {"type": "integer"},
                {
                    "type": "object",
                    "properties": {
                        "foo": {"enum": [2]}
                    }
                }
            ]
        }

        validator = PandelVisitor(schema)
        data = {"foo": 1}
        for instance in wrap_in_pandas(data):
            errors = list(validator.iter_errors(instance))
            self.assertEqual(len(errors), 1)
            e = errors[0]

            self.assertEqual(e.validator, "type")
            self.assertEqual(e.validator_value, schema["type"])
            npt.assert_array_equal(e.instance, instance)
            self.assertEqual(e.schema, schema)
            self.assertIsNone(e.parent)

            self.assertEqual(e.path, deque([]))
            self.assertEqual(e.relative_path, deque([]))
            self.assertEqual(e.absolute_path, deque([]))

            self.assertEqual(e.schema_path, deque(["type"]))
            self.assertEqual(e.relative_schema_path, deque(["type"]))
            self.assertEqual(e.absolute_schema_path, deque(["type"]))

            self.assertEqual(len(e.context), 2)

            e1, e2 = sorted_errors(e.context)

            self.assertEqual(e1.validator, "type")
            self.assertEqual(e1.validator_value, schema["type"][0]["type"])
            npt.assert_array_equal(e1.instance, instance)
            self.assertEqual(e1.schema, schema["type"][0])
            self.assertIs(e1.parent, e)

            self.assertEqual(e1.path, deque([]))
            self.assertEqual(e1.relative_path, deque([]))
            self.assertEqual(e1.absolute_path, deque([]))

            self.assertEqual(e1.schema_path, deque([0, "type"]))
            self.assertEqual(e1.relative_schema_path, deque([0, "type"]))
            self.assertEqual(
                e1.absolute_schema_path, deque(["type", 0, "type"]))

            self.assertFalse(e1.context)

            self.assertEqual(e2.validator, "enum")
            self.assertEqual(e2.validator_value, [2])
            npt.assert_array_equal(e2.instance, 1)
            self.assertEqual(e2.schema, {u"enum": [2]})
            self.assertIs(e2.parent, e)

            self.assertEqual(e2.path, deque(["foo"]))
            self.assertEqual(e2.relative_path, deque(["foo"]))
            self.assertEqual(e2.absolute_path, deque(["foo"]))

            self.assertEqual(
                e2.schema_path, deque([1, "properties", "foo", "enum"]),
            )
            self.assertEqual(
                e2.relative_schema_path, deque(
                    [1, "properties", "foo", "enum"]),
            )
            self.assertEqual(
                e2.absolute_schema_path,
                deque(["type", 1, "properties", "foo", "enum"]),
            )

            self.assertFalse(e2.context)

    def test_single_nesting(self):
        data = {"foo": 2, "bar": [1], "baz": 15, "quux": "spam"}
        schema = {
            "properties": {
                "foo": {"type": "string"},
                "bar": {"minItems": 2},
                "baz": {"maximum": 10, "enum": [2, 4, 6, 8]},
            }
        }

        validator = PandelVisitor(schema)
        for instance in wrap_in_pandas(data):
            errors = validator.iter_errors(instance)
            errors = sorted_errors(errors)
            try:
                e1, e2, e3, e4 = errors
            except ValueError as ex:
                print(list(sorted_errors(validator.iter_errors(instance))))
                raise ex

            self.assertEqual(e1.path, deque(["bar"]))
            self.assertEqual(e2.path, deque(["baz"]))
            self.assertEqual(e3.path, deque(["baz"]))
            self.assertEqual(e4.path, deque(["foo"]))

            self.assertEqual(e1.relative_path, deque(["bar"]))
            self.assertEqual(e2.relative_path, deque(["baz"]))
            self.assertEqual(e3.relative_path, deque(["baz"]))
            self.assertEqual(e4.relative_path, deque(["foo"]))

            self.assertEqual(e1.absolute_path, deque(["bar"]))
            self.assertEqual(e2.absolute_path, deque(["baz"]))
            self.assertEqual(e3.absolute_path, deque(["baz"]))
            self.assertEqual(e4.absolute_path, deque(["foo"]))

            self.assertEqual(e1.validator, "minItems")
            self.assertEqual(e2.validator, "enum")
            self.assertEqual(e3.validator, "maximum")
            self.assertEqual(e4.validator, "type")

    #@unittest.skip("For Draft3Validator only!")
    def test_multiple_nesting(self):
        instance = [1, {"foo": 2, "bar": {"baz": [1]}}, "quux"]
        schema = {
            "$schema": "http://json-schema.org/draft-03/schema#",
            "type": "string",
            "items": {
                "type": ["string", "object"],
                "properties": {
                    "foo": {"enum": [1, 3]},
                    "bar": {
                        "type": "array",
                        "properties": {
                            "bar": {"required": True},
                            "baz": {"minItems": 2},
                        }
                    }
                }
            }
        }

        validator = PandelVisitor(schema)
        errors = validator.iter_errors(instance)
        e1, e2, e3, e4, e5, e6 = sorted_errors(errors)

        self.assertEqual(e1.path, deque([]))
        self.assertEqual(e2.path, deque([0]))
        self.assertEqual(e3.path, deque([1, "bar"]))
        self.assertEqual(e4.path, deque([1, "bar", "bar"]))
        self.assertEqual(e5.path, deque([1, "bar", "baz"]))
        self.assertEqual(e6.path, deque([1, "foo"]))

        self.assertEqual(e1.schema_path, deque(["type"]))
        self.assertEqual(e2.schema_path, deque(["items", "type"]))
        self.assertEqual(
            list(e3.schema_path), ["items", "properties", "bar", "type"],
        )
        self.assertEqual(
            list(e4.schema_path),
            ["items", "properties", "bar", "properties", "bar", "required"],
        )
        self.assertEqual(
            list(e5.schema_path),
            ["items", "properties", "bar", "properties", "baz", "minItems"]
        )
        self.assertEqual(
            list(e6.schema_path), ["items", "properties", "foo", "enum"],
        )

        self.assertEqual(e1.validator, "type")
        self.assertEqual(e2.validator, "type")
        self.assertEqual(e3.validator, "type")
        self.assertEqual(e4.validator, "required")
        self.assertEqual(e5.validator, "minItems")
        self.assertEqual(e6.validator, "enum")

    def test_recursive(self):
        schema = {
            "definitions": {
                "node": {
                    "anyOf": [{
                        "type": "object",
                        "required": ["name", "children"],
                        "properties": {
                            "name": {
                                "type": "string",
                            },
                            "children": {
                                "type": "object",
                                "patternProperties": {
                                    "^.*$": {
                                        "$ref": "#/definitions/node",
                                    },
                                },
                            },
                        },
                    }],
                },
            },
            "type": "object",
            "required": ["root"],
            "properties": {
                "root": {"$ref": "#/definitions/node"},
            }
        }

        instance = {
            "root": {
                "name": "root",
                "children": {
                    "a": {
                        "name": "a",
                        "children": {
                            "ab": {
                                "name": "ab",
                                # missing "children"
                            }
                        }
                    },
                },
            },
        }
        validator = PandelVisitor(schema)

        e, = validator.iter_errors(instance)
        self.assertEqual(e.absolute_path, deque(["root"]))
        self.assertEqual(
            e.absolute_schema_path, deque(["properties", "root", "anyOf"]),
        )

        e1, = e.context
        self.assertEqual(e1.absolute_path, deque(["root", "children", "a"]))
        self.assertEqual(
            e1.absolute_schema_path, deque(
                [
                    "properties",
                    "root",
                    "anyOf",
                    0,
                    "properties",
                    "children",
                    "patternProperties",
                    "^.*$",
                    "anyOf",
                ],
            ),
        )

        e2, = e1.context
        self.assertEqual(
            e2.absolute_path, deque(
                ["root", "children", "a", "children", "ab"],
            ),
        )
        self.assertEqual(
            e2.absolute_schema_path, deque(
                [
                    "properties",
                    "root",
                    "anyOf",
                    0,
                    "properties",
                    "children",
                    "patternProperties",
                    "^.*$",
                    "anyOf",
                    0,
                    "properties",
                    "children",
                    "patternProperties",
                    "^.*$",
                    "anyOf"
                ],
            ),
        )

    def test_additionalProperties(self):
        data = {"bar": "bar", "foo": 2}
        schema = {
            "additionalProperties": {"type": "integer", "minimum": 5}
        }

        validator = PandelVisitor(schema)

        for instance in wrap_in_pandas(data):
            errors = validator.iter_errors(instance)
            e1, e2 = sorted_errors(errors)

            self.assertEqual(e1.path, deque(["bar"]))
            self.assertEqual(e2.path, deque(["foo"]))

            self.assertEqual(e1.validator, "type")
            self.assertEqual(e2.validator, "minimum")

    def test_patternProperties(self):
        data = {"bar": 1, "foo": 2}
        schema = {
            "patternProperties": {
                "bar": {"type": "string"},
                "foo": {"minimum": 5}
            }
        }

        validator = PandelVisitor(schema)

        for instance in wrap_in_pandas(data):
            errors = validator.iter_errors(instance)
            try:
                e1, e2 = sorted_errors(errors)
            except ValueError as ex:
                print(list(sorted_errors(validator.iter_errors(instance))))
                raise ex

            self.assertEqual(e1.path, deque(["bar"]))
            self.assertEqual(e2.path, deque(["foo"]))

            self.assertEqual(e1.validator, "type")
            self.assertEqual(e2.validator, "minimum")

    def test_additionalItems(self):
        data = ["foo", 1]
        schema = {
            "$schema": "http://json-schema.org/draft-03/schema#",
            "items": [],
            "additionalItems": {"type": "integer", "minimum": 5}
        }

        validator = PandelVisitor(schema)

        for instance in wrap_in_pandas(data):
            errors = validator.iter_errors(instance)
            e1, e2 = sorted_errors(errors)

            self.assertEqual(e1.path, deque([0]))
            self.assertEqual(e2.path, deque([1]))

            if not isinstance(instance, np.ndarray):
                # numpy-arrays have column-types so str+int-->str and both
                # errors are type-errors.
                self.assertSequenceEqual(
                    [e.validator for e in (e1, e2)], ("type", "minimum"), (e1, e2))

    def test_additionalItems_with_items(self):
        data = ["foo", "bar", 1]
        schema = {
            "items": [{}],
            "additionalItems": {"type": "integer", "minimum": 5}
        }

        validator = PandelVisitor(schema)

        for instance in wrap_in_pandas(data):
            errors = validator.iter_errors(instance)
            e1, e2 = sorted_errors(errors)

            self.assertEqual(e1.path, deque([1]))
            self.assertEqual(e2.path, deque([2]))

            if not isinstance(instance, np.ndarray):
                # numpy-arrays have column-types so str+int-->str and both
                # errors are type-errors.
                self.assertSequenceEqual(
                    [e.validator for e in (e1, e2)], ("type", "minimum"), (e1, e2))


class ValidatorTestMixin(object):

    def setUp(self):
        self.instance = mock.Mock()
        self.schema = {}
        self.resolver = mock.Mock()
        self.validator = self.validator_class(self.schema)

    def test_valid_instances_are_valid(self):
        errors = iter([])

        with mock.patch.object(                                     # @UndefinedVariable
            self.validator, "iter_errors", return_value=errors,
        ):
            self.assertTrue(
                self.validator.is_valid(self.instance, self.schema)
            )

    def test_invalid_instances_are_not_valid(self):
        errors = iter([mock.Mock()])

        with mock.patch.object(                                     # @UndefinedVariable
            self.validator, "iter_errors", return_value=errors,
        ):
            self.assertFalse(
                self.validator.is_valid(self.instance, self.schema)
            )

    def test_non_existent_properties_are_ignored(self):
        instance, my_property, my_value = mock.Mock(), mock.Mock(), mock.Mock()
        validate(instance=instance, schema={my_property: my_value})

    def test_it_creates_a_ref_resolver_if_not_provided(self):
        self.assertIsInstance(self.validator.resolver, RefResolver)

    def test_it_delegates_to_a_ref_resolver(self):
        resolver = RefResolver("", {})
        schema = {"$ref": mock.Mock()}

        @contextmanager
        def resolving():
            yield {"type": "integer"}

        with mock.patch.object(resolver, "resolve") as resolve:
            resolve.return_value = "url", {"type": "integer"}
            with self.assertRaises(ValidationError):
                self.validator_class(schema, resolver=resolver).validate(None)

        resolve.assert_called_once_with(schema["$ref"])

    def test_is_type_is_true_for_valid_type(self):
        self.assertTrue(self.validator.is_type("foo", "string"))

    def test_is_type_is_false_for_invalid_type(self):
        self.assertFalse(self.validator.is_type("foo", "array"))

    def test_is_type_evades_bool_inheriting_from_int(self):
        self.assertFalse(self.validator.is_type(True, "integer"))
        self.assertFalse(self.validator.is_type(True, "number"))

    def test_is_type_raises_exception_for_unknown_type(self):
        with self.assertRaises(UnknownType):
            self.validator.is_type("foo", object())


class TestDraft3lValidator(ValidatorTestMixin, unittest.TestCase):
    validator_class = PandelVisitor

    def setUp(self):
        super(TestDraft3lValidator, self).setUp()
        self.validator = PandelVisitor({
            "$schema": "http://json-schema.org/draft-03/schema#"
        })

    #@unittest.skip("For Draft3Validator only!")
    def test_is_type_is_true_for_any_type(self):
        self.assertTrue(self.validator.is_valid(mock.Mock(), {"type": "any"}))

    #@unittest.skip("For Draft3Validator only!")
    def test_is_type_does_not_evade_bool_if_it_is_being_tested(self):
        self.assertTrue(self.validator.is_type(True, "boolean"))
        self.assertTrue(self.validator.is_valid(True, {"type": "any"}))

    #@unittest.skip("The schema below in invalid under Draft3/4, but original test had averted meta-validation.")
    def test_non_string_custom_types(self):
        schema = {
            "$schema": "http://json-schema.org/draft-03/schema#",
            'type': [None]
        }
        cls = self.validator_class(
            schema, types={None: type(None)}, skip_meta_validation=True)
        cls.validate(None, schema)

        schema = {
            "$schema": "http://json-schema.org/draft-03/schema#",
            'type': 'some'
        }
        types = {
            'object': dict,
            'some': pd.DataFrame,
        }
        cls = self.validator_class(
            schema, types=types, skip_meta_validation=True)
        cls.validate(pd.DataFrame(), schema)


class TestDraft4Validator(ValidatorTestMixin, unittest.TestCase):
    validator_class = PandelVisitor


class TestSchemaIsChecked(unittest.TestCase):  # Was: class TestValidate

    def test_draft3_validator_is_chosen(self):
        pv = PandelVisitor({})
        schema = {"$schema": "http://json-schema.org/draft-03/schema#"}
        with mock.patch.object(pv, "iter_errors", return_value=()) as chk_schema:
            pv.validate({}, schema)
            chk_schema.assert_called_once_with({}, schema)

        # Make sure it works without the empty fragment
        pv = PandelVisitor({})
        schema = {"$schema": "http://json-schema.org/draft-03/schema"}
        with mock.patch.object(pv, "iter_errors", return_value=()) as chk_schema:
            pv.validate({}, schema)
            chk_schema.assert_called_once_with({}, schema)

    def test_draft4_validator_is_chosen(self):
        pv = PandelVisitor({})
        schema = {"$schema": "http://json-schema.org/draft-04/schema#"}
        with mock.patch.object(pv, "iter_errors", return_value=()) as chk_schema:
            pv.validate({}, schema)
            chk_schema.assert_called_once_with({}, schema)

    def test_draft4_validator_is_the_default(self):
        pv = PandelVisitor({})
        with mock.patch.object(pv, "iter_errors", return_value=()) as chk_schema:
            pv.validate({}, {})
            chk_schema.assert_called_once_with({}, {})


def sorted_errors(errors):
    def key(error):
        return (
            [str(e) for e in error.path],
            [str(e) for e in error.schema_path]
        )
    return sorted(errors, key=key)



#! python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2015 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Hierarchical string-like objects useful for indexing, that can be rename/relocated at a later stage.

.. autosummary::

    Pstep
    pmods_from_tuples
    Pmod

*Example*::

    >>> from pandalone.mappings import pmods_from_tuples

    >>> pmods = pmods_from_tuples([
    ...     ('',         'deeper/ROOT'),
    ...     ('/abc',     'ABC'),
    ...     ('/abc/foo', 'BAR'),
    ... ])
    >>> p = pmods.step()
    >>> p.abc.foo
    `BAR`
    >>> p._paths()
    ['deeper/ROOT/ABC/BAR']

- TODO: Implements "anywhere" pmods(`//`).
"""

from __future__ import division, unicode_literals

from collections import OrderedDict
from copy import copy
import logging
import re

from past.builtins import basestring

import functools as ft
from pandalone import utils
from pandalone.pandata import (
    iter_jsonpointer_parts_relaxed, JSchema, unescape_jsonpointer_part,
    escape_jsonpointer_part)


__commit__ = ""

log = logging.getLogger(__name__)


class Pmod(object):

    r"""
    A path-step mapping forming the pmods-hierarchy.

    - The :term:`pmods` denotes the hierarchy of all :term:`mappings`,
      that either *rename* or *relocate* path-steps.

    - A single :term:`mapping` transforms an "origin" path to
      a "destination" one (also called as "from" and "to" paths).

    - A mapping always transforms the *final* path-step, like that::

        FROM_PATH       TO_PATH       RESULT_PATH
        ---------       -------       -----------
        /rename/path    foo       --> /rename/foo        ## renaming
        /relocate/path  foo/bar   --> /relocate/foo/bar  ## relocation
        ''              a/b/c     --> /a/b/c             ## Relocate all paths.
        /               a/b/c     --> /a/b/c             ## Relocates 1st "empty-str" step.

    - The :term:`pmod` is the mapping of that single path-step.

    - It is possible to match fully on path-steps using regular-expressions,
      and then to use any captured-groups from the *final* step into
      the mapped value::

        (/all(.*)/path, foo)   + all_1/path --> /all_1/foo
                               + all_XYZ    --> /all_XYZ        ## no change
        (/all(.*)/path, foo\1) + all_1/path --> /all_1/foo_1

      If more than one regex match, they are merged in the order declared
      (the latest one overrides a previous one).

    - Any exact child-name matches are applied and merged after regexs.

    - Use :func:`pmods_from_tuples()` to construct the pmods-hierarchy.

    - The pmods are used internally by class:`Pstep` to correspond
      the component-paths of their input & output onto the actual
      value-tree paths.

    :ivar str _alias:          (optional) the mapped-name of the pstep for
                               this pmod
    :ivar dict _steps:         {original_name --> pmod}
    :ivar OrderedDict _regxs:  {regex_on_originals --> pmod}


    Example:

    .. Note::
        Do not manually construct instances from this class!
        To construct a hierarchy use the :func:`pmods_from_tuples()`.

    You can either use it for massively map paths, either for *renaming* them::

        >>> pmods = pmods_from_tuples([
        ...         ('/a',           'A'),
        ...         ('/~b.*',        r'BB\g<0>'),  ## Previous match.
        ...         ('/~b.*/~c.(.*)', r'W\1ER'),    ## Capturing-group(1)
        ... ])
        >>> pmods.map_paths(['/a', '/a/foo'])     ## 1st rule
        ['/A', '/A/foo']

        >>> pmods.map_path('/big/stuff')          ## 2nd rule
        '/BBbig/stuff'

        >>> pmods.map_path('/born/child')         ## 2nd & 3rd rule
        '/BBborn/WildER'


    or to *relocate* them::

        >>> pmods = pmods_from_tuples([
        ...         ('/a',           'A/AA'),
        ...         ('/~b.*/~c(.*)',  r'../C/\1'),
        ...         ('/~b.*/~.*/~r.*', r'/\g<0>'),
        ... ])
        >>> pmods.map_paths(['/a/foo', '/big/child', '/begin/from/root'])
        ['/A/AA/foo', '/big/C/hild', '/root']


    Here is how you relocate "root"
    (notice that the `''` path is the root)::

        >>> pmods = pmods_from_tuples([('', '/NEW/ROOT')])
        >>> pmods.map_paths(['/a/foo', ''])
        ['/NEW/ROOT/a/foo', '/NEW/ROOT']

    """

    __slots__ = ['_alias', '_steps', '_regxs']

    def __init__(self, _alias=None, _steps={}, _regxs={}):
        """
        Args passed only for testing, remember `_regxs` to be (k,v) tuple-list!

        .. Note:: Volatile arg-defaults (empty dicts) are knowingly used ,
            to preserve memory; should never append in them!

        """
        self._alias = _alias
        self._steps = _steps
        if _regxs:
            self._regxs = OrderedDict(
                (re.compile(k), v) for k, v in _regxs)
        else:
            self._regxs = _regxs

    def step(self, pname='', alias=None):
        """
        Create a new :class:`Pstep` having as mappings this pmod.

        If no `pname` specified, creates a *root* pstep.

        Delegates to :meth:`Pstep.__new__()`.
        """
        return Pstep(pname, _proto_or_pmod=self, alias=alias)

    def _append_into_steps(self, key):
        """
        Inserts a child-mappings into `_steps` dict.

        :param str key:    the step-name to add
        """

        cpmod = None
        d = self._steps
        if not d:
            self._steps = d = {}  # Do not modify init-defaults.
        else:
            cpmod = d.get(key)
        if not cpmod:
            d[key] = cpmod = Pmod()

        return cpmod

    def _append_into_regxs(self, key):
        """
        Inserts a child-mappings into `_steps` dict.

        :param str key:    the regex-pattern to add
        """
        key = re.compile(key)
        cpmod = None
        d = self._regxs
        if not d:
            self._regxs = d = OrderedDict()  # Do not modify init-defaults.
        else:
            cpmod = d.get(key)
            if cpmod:
                # Remove it, to append it at then end.
                del d[key]
        if not cpmod:
            cpmod = Pmod()
        d[key] = cpmod

        return cpmod

    def _override_steps(self, other):
        """
        Override this pmod's '_steps' dict with other's, recursively.

        Same as :meth:`_override_regxs()` but without caring for order.
        """

        opmods = other._steps
        if opmods:
            spmods = self._steps
            if spmods:
                # Like ``spmods.copy().update()`` but
                # recursive `_merge()` on common items.
                #
                spmods = spmods.copy()
                for name, opmod in opmods.items():
                    spmod = spmods.get(name)
                    if spmod:
                        opmod = spmod._merge(opmod)
                    spmods[name] = opmod  # Share other-pmod if not mine.
                opmods = spmods

            # Share other dict if self hadn't its own.
            self._steps = opmods

    def _override_regxs(self, other):
        """
        Override this pmod's `_regxs` dict with other's, recursively.

        - It may "share" (crosslink) the dict and/or its child-pmods
          between the two pmod args (`self` and `other`).
        - No dict is modified (apart from self, which must have been cloned
          previously by :meth:`Pmod._merge()`), to avoid side-effects
          in case they were "shared".
        - It preserves dict-ordering so that `other` order takes precedence
          (its elements are the last ones).

        :param Pmod self:    contains the dict that would be overridden
        :param Pmod other:   contains the dict with the overrides
        """

        opmods = other._regxs
        if opmods:
            spmods = self._regxs
            if spmods:
                # Like ``spmods.copy().update()`` but
                # with recursive `_merge()` on common items,
                # and preserve order.
                #
                opairs = []
                for name, opmod in opmods.items():
                    spmod = spmods.get(name)
                    if spmod:
                        mpmod = spmod._merge(opmod)
                    else:
                        mpmod = opmod  # Share other-pmod.
                    opairs.append((name, mpmod))

                okeys = opmods.keys()
                spairs = [(name, spmod)  # Share self-pmod.
                          for name, spmod in spmods.items()
                          if name not in okeys]

                opmods = type(spmods)(spairs + opairs)

            # Share other dict if self hadn't its own.
            self._regxs = opmods

    def _merge(self, other):
        """
        Clone and override all its props with props from other-pmod, recursively.

        Although it does not modify this, the `other` or their children pmods,
        it may "share" (crosslink) them, so pmods MUST NOT be modified later.

        :param Pmod other: contains the dicts with the overrides
        :return:           the cloned merged pmod
        :rtype:            Pmod

        Examples:

        Look how `_steps` are merged::

            >>> pm1 = Pmod(_alias='pm1', _steps={
            ...     'a':Pmod(_alias='A'), 'c':Pmod(_alias='C')})
            >>> pm2 = Pmod(_alias='pm2', _steps={
            ...     'b':Pmod(_alias='B'), 'a':Pmod(_alias='AA')})
            >>> pm = pm1._merge(pm2)
            >>> sorted(pm._steps.keys())
            ['a', 'b', 'c']


        And here it is `_regxs` merging, which preserves order::

            >>> pm1 = Pmod(_alias='pm1',
            ...            _regxs=[('d', Pmod(_alias='D')),
            ...                    ('a', Pmod(_alias='A')),
            ...                    ('c', Pmod(_alias='C'))])
            >>> pm2 = Pmod(_alias='pm2',
            ...            _regxs=[('b', Pmod(_alias='BB')),
            ...                    ('a', Pmod(_alias='AA'))])

            >>> pm1._merge(pm2)
            pmod('pm2', OrderedDict([(re.compile('d'), pmod('D')),
                       (re.compile('c'), pmod('C')),
                       (re.compile('b'), pmod('BB')),
                       (re.compile('a'), pmod('AA'))]))

            >>> pm2._merge(pm1)
            pmod('pm1', OrderedDict([(re.compile('b'), pmod('BB')),
                        (re.compile('d'), pmod('D')),
                        (re.compile('a'), pmod('A')),
                        (re.compile('c'), pmod('C'))]))
        """
        self = copy(self)
        if other._alias is not None:
            self._alias = other._alias
        self._override_steps(other)
        self._override_regxs(other)

        return self

    def _match_regxs(self, cstep):
        """Return (pmod, regex.match) for those child-pmods matching `cstep`."""

        return [(rpmod, match)
                for rpmod, match
                in ((rpmod, utils.fullmatch(regex, cstep))
                    for regex, rpmod
                    in self._regxs.items())
                if match]

    def descend(self, cstep):
        r"""
        Return child-pmod with merged any exact child with all matched regexps, along with its alias regex-expaned.

        :param str cstep:   the child path-step cstep of the pmod to return
        :return:            the merged-child pmod, along with the alias;
                            both might be None, if nothing matched, or no alias.
        :rtype:             tuple(Pmod, str)

        Example::

            >>> pm = Pmod(
            ...     _steps={'a': Pmod(_alias='A')},
            ...     _regxs=[('a\w*', Pmod(_alias='AWord')),
            ...              ('a(\d*)', Pmod(_alias=r'A_\1')),
            ...    ])
            >>> pm.descend('a')
            (pmod('A'), 'A')

            >>> pm.descend('abc')
            (pmod('AWord'), 'AWord')

            >>> pm.descend('a12')
            (pmod('A_\\1'), 'A_12')

            >>> pm.descend('BAD')
            (None, None)


        Notice how children of regexps are merged together::

            >>> pm = Pmod(
            ...     _steps={'a':
            ...        Pmod(_alias='A', _steps={1: 11})},
            ...     _regxs=[
            ...        (r'a\w*', Pmod(_alias='AWord',
            ...                      _steps={2: Pmod(_alias=22)})),
            ...        (r'a\d*', Pmod(_alias='ADigit',
            ...                     _steps={3: Pmod(_alias=33)})),
            ...    ])
            >>> sorted(pm.descend('a')[0]._steps)    ## All children and regexps match.
            [1, 2, 3]

            >>> pm.descend('aa')[0]._steps           ## Only 'a\w*' matches.
            {2: pmod(22)}

            >>> sorted(pm.descend('a1')[0]._steps )  ## Both regexps matches.
            [2, 3]

        So it is possible to say::

            >>> pm.descend('a1')[0].alias(2)
            22
            >>> pm.descend('a1')[0].alias(3)
            33
            >>> pm.descend('a1')[0].descend('BAD')
            (None, None)
            >>> pm.descend('a$')
            (None, None)

        but it is better to use :meth:`map_path()` for this.
        """
        alias = None

        cpmod = self._steps.get(cstep)
        pmods = self._match_regxs(cstep)

        if cpmod and cpmod._alias is not None:
            alias = cpmod._alias
        else:
            for rpmod, match in reversed(pmods):
                if rpmod._alias is not None:
                    alias = match.expand(rpmod._alias)
                    break
        pmods = [pmod for pmod, _ in pmods]
        if cpmod:
            pmods.append(cpmod)

        if pmods:
            return (ft.reduce(Pmod._merge, pmods), alias)
        return (None, None)

    def alias(self, cstep):
        """
        Like :meth:`descend()` but without merging child-pmods.

        :return: the expanded alias from child/regexs or None
        """
        cpmod = self._steps.get(cstep)
        if cpmod and cpmod._alias is not None:
            return cpmod._alias

        pmods = self._match_regxs(cstep)

        for rpmod, match in reversed(pmods):
            if rpmod._alias is not None:
                return match.expand(rpmod._alias)

    def map_path(self, path):
        r"""
        Maps a '/rooted/path' using all aliases while descending its child pmods.

        It uses any aliases on all child pmods if found.

        :param str path: a rooted path to transform
        :return:         the rooted mapped path or '/' if path was '/'
        :rtype:          str or None

        Examples::

            >>> pmods = pmods_from_tuples([
            ...         ('/a',              'A/AA'),
            ...         ('/~a(\\w*)',       r'BB\1'),
            ...         ('/~a\\w*/~d.*',     r'D \g<0>'),
            ...         ('/~a(\\d+)',       r'C/\1'),
            ...         ('/~a(\\d+)/~(c.*)', r'CC-/\1'), # The 1st group is ignored!
            ...         ('/~a\\d+/~e.*',     r'/newroot/\g<0>'), # Rooted mapping.
            ... ])

            >>> pmods.map_path('/a')
            '/A/AA'

            >>> pmods.map_path('/a_hi')
            '/BB_hi'

            >>> pmods.map_path('/a12')
            '/C/12'

            >>> pmods.map_path('/a12/etc')
            '/newroot/etc'

        Notice how children from *all* matching prior-steps are merged::

            >>> pmods.map_path('/a12/dow')
            '/C/12/D dow'
            >>> pmods.map_path('/a12/cow')
            '/C/12/CC-/cow'


        To map *root* use '' which matches before the 1st slash('/')::

            >>> pmods = pmods_from_tuples([('', 'New/Root'),])  ## Relative
            >>> pmods
            pmod({'': pmod('New/Root')})

            >>> pmods.map_path('/for/plant')
            'New/Root/for/plant'

            >>> pmods_from_tuples([('', '/New/Root'),]).map_path('/for/plant')
            '/New/Root/for/plant'

        .. Note::
            Using slash('/') for "from" path will NOT map *root*::

                >>> pmods = pmods_from_tuples([('/', 'New/Root'),])
                >>> pmods
                pmod({'': pmod({'': pmod('New/Root')})})

                >>> pmods.map_path('/for/plant')
                '/for/plant'

                >>> pmods.map_path('//for/plant')
                '/New/Root/for/plant'

                '/root'

        but '' always remains unchanged (whole document)::

            >>> pmods.map_path('')
            ''

        """
        is_folder = len(path) > 1 and path.endswith('/')
        if is_folder:
            path = path[:-1]

        steps = tuple(iter_jsonpointer_parts_relaxed(path))
        if self._alias is None:
            nsteps = ()
        else:
            nsteps = tuple(iter_jsonpointer_parts_relaxed(self._alias))

        if steps:
            pmod = self
            # Separate last-step from loop below, since
            #    merged child-pmods in `descend` are not needed.
            #
            for step in steps[:-1]:
                if pmod:
                    pmod, alias = pmod.descend(step)
                if alias is not None:
                    if alias.startswith('.'):
                        nsteps += (step, )
                    step = alias
                # XXX: Monkey business here.
                if len(step) > 1 and step.endswith('/'):
                    step = step[:-1]
                nsteps += tuple(iter_jsonpointer_parts_relaxed(step))

            final_step = steps[-1]
            if pmod:
                alias = pmod.alias(final_step)
                if alias is not None:
                    if alias.startswith('.'):
                        nsteps += (final_step, )
                    final_step = alias
            # XXX: Monkey business here.
            is_folder = len(final_step) > 1 and final_step.endswith('/')
            if is_folder:
                final_step = final_step[:-1]
            nsteps += tuple(iter_jsonpointer_parts_relaxed(final_step))

        npath = _join_paths(*nsteps)

        if is_folder:
            path += '%s/' % path

        return npath

    def map_paths(self, paths):
        return [self.map_path(p) for p in paths]

    def __repr__(self):
        args = [repr(a)
                for a in [self._alias, self._steps, self._regxs]
                if a or a == '']

        args = ', '.join(args)
        return 'pmod({})'.format(args)

    def __eq__(self, o):
        try:
            return (self._alias, self._steps, self._regxs) == (o._alias, o._steps, o._regxs)
        except:
            return False


def pmods_from_tuples(pmods_tuples):
    """
    Turns a list of 2-tuples into a *pmods* hierarchy.

    - Each tuple defines the renaming-or-relocation of the *final* part
      of some component path onto another one into value-trees, such as::

          (/rename/path, foo)          --> rename/foo
          (relocate/path, foo/bar)    --> relocate/foo/bar


    - The "from" path may be:
      - relative,
      - absolute(starting with `/`), or
      - "anywhere"(starting with `//`).

    - In case a "step" in the "from" path starts with tilda(`~`),
      it is assumed to be a regular-expression, and it is removed from it.
      The "to" path can make use of any "from" capture-groups::

          ('/~all(.*)/path', 'foo')
          ('~some[\d+]/path', 'foo\1')
          ('//~all(.*)/path', 'foo')



    :param list(tuple(str, str) pmods_tuples:
    :return: a root pmod
    :rtype:  Pmod


    Example::

        >>> pmods_from_tuples([
        ...     ('/a', 'A1/A2'),
        ...     ('/a/b', 'B'),
        ... ])
        pmod({'': pmod({'a': pmod('A1/A2', {'b': pmod('B')})})})

        >>> pmods_from_tuples([
        ...     ('/~a*', 'A1/A2'),
        ...     ('/a/~b[123]', 'B'),
        ... ])
        pmod({'': pmod({'a':
                pmod(OrderedDict([(re.compile('b[123]'), pmod('B'))]))},
                     OrderedDict([(re.compile('a*'), pmod('A1/A2'))]))})


    This is how you map *root*::

        >>> pmods = pmods_from_tuples([
        ...     ('', 'relative/Root'),        ## Make all paths relatives.
        ...     ('/a/b', '/Rooted/B'),        ## But map `b` would be "rooted".
        ... ])
        >>> pmods
        pmod({'':
                pmod('relative/Root',
                        {'a': pmod({'b':
                                pmod('/Rooted/B')})})})

        >>> pmods.map_path('/a/c')
        'relative/Root/a/c'

        >>> pmods.map_path('/a/b')
        '/Rooted/B'


    But note that '/' maps the 1st "empty-str" step after root::

        >>> pmods_from_tuples([
        ...     ('/', 'New/Root'),
        ... ])
        pmod({'': pmod({'': pmod('New/Root')})})


    TODO: Implement "anywhere" matches.
    """
    root = Pmod()
    for i, (f, t) in enumerate(pmods_tuples):
        if (f, t) == ('', '') or f is None or t is None:
            msg = 'pmod-tuple #%i of %i: Invalid "from-to" tuple (%r, %r).'
            log.warning(msg, i + 1, len(pmods_tuples), f, t)
            continue

        pmod = root
        for srcstep in iter_jsonpointer_parts_relaxed(f):
            is_regex = srcstep.startswith('~')
            if is_regex:
                pmod = pmod._append_into_regxs(srcstep[1:])
            else:
                pmod = pmod._append_into_steps(srcstep)

        pmod._alias = t

    return root


def _append_step(steps, step):
    """
    Joins `step` at the right of `steps`, respecting '/', '..', '.', ''.

    :param tuple steps:  where to append into
                         ("absolute" when 1st-element is '')
    :param str step:     what to append
                         (may be: ``'foo', '.', '..', ''``)
    :rtype:  tuple

    .. Note::
        The empty-string('') is the "root" for both `steps` and `step`.
        An empty-tuple `steps` is considered "relative", equivalent to dot(`.`).


    Example::

        >>> _append_step((), 'a')
        ('a',)

        >>> _append_step(('a', 'b'), '..')
        ('a',)

        >>> _append_step(('a', 'b'), '.')
        ('a', 'b')


    Not that an "absolute" path has the 1st-step empty(`''`),
    (so the previous paths above were all "relative")::

        >>> _append_step(('a', 'b'), '')
        ('',)

        >>> _append_step(('',), '')
        ('',)

        >>> _append_step((), '')
        ('',)


    Dot-dots preserve "relative" and "absolute" paths, respectively,
    and hence do not coalesce when at the left::

        >>> _append_step(('',), '..')
        ('',)

        >>> _append_step(('',), '.')
        ('',)

        >>> _append_step(('a',), '..')
        ()

        >>> _append_step((), '..')
        ('..',)

        >>> _append_step(('..',), '..')
        ('..', '..')

        >>> _append_step((), '.')
        ()



    Single-dots('.') just dissappear::

        >>> _append_step(('.',), '.')
        ()

        >>> _append_step(('.',), '..')
        ('..',)

    """
    assert isinstance(steps, tuple), (steps, step)
    assert not step or isinstance(step, basestring), (steps, step)

    if step == '':
        return ('',)

    _last_pair_choices = {
        ('.',): (),
        ('..',): ('..',),
        ('.', '.'): (),
        ('.', '..',): ('..',),
        ('..', '.'): ('..',),
        ('..', '..'): ('..', '..'),
        ('', '.'): ('',),
        ('', '..'): ('',),
    }

    try:
        last_pair = steps[-1:] + (step,)
        steps = steps[:-1] + _last_pair_choices[last_pair]
    except KeyError:
        if step == '.':
            pass
        elif step == '..':
            steps = steps[:-1]
        else:
            steps += (step,)

    return steps


def _join_paths(*steps):
    """
    Joins all path-steps in a single string, respecting ``'/', '..', '.', ''``.

    :param str steps:  single json-steps, from left to right
    :rtype: str

    .. Note::
        If you use :func:`iter_jsonpointer_parts_relaxed()` to generate
        path-steps, the "root" is signified by the empty(`''`) step;
        not the slash(`/`)!

        Hence a lone slash(`/`) gets splitted to an empty step after "root"
        like that: ``('', '')``, which generates just "root"(`''`).

        Therefore a "folder" (i.e. `some/folder/`) when splitted equals
        ``('some', 'folder', '')``, which results again in the "root"(`''`)!


    Examples::

        >>> _join_paths('r', 'a', 'b')
        'r/a/b'

        >>> _join_paths('', 'a', 'b', '..', 'bb', 'cc')
        '/a/bb/cc'

        >>> _join_paths('a', 'b', '.', 'c')
        'a/b/c'


    An empty-step "roots" the remaining path-steps::

        >>> _join_paths('a', 'b', '', 'r', 'aa', 'bb')
        '/r/aa/bb'


    All `steps` have to be already "splitted"::

        >>> _join_paths('a', 'b', '../bb')
        'a/b/../bb'


    Dot-doting preserves "relative" and "absolute" paths, respectively::

        >>> _join_paths('..')
        '..'

        >>> _join_paths('a', '..')
        '.'

        >>> _join_paths('a', '..', '..', '..')
        '../..'

        >>> _join_paths('', 'a', '..', '..')
        ''


    Some more special cases::

        >>> _join_paths('..', 'a')
        '../a'

        >>> _join_paths('', '.', '..', '..')
        ''

        >>> _join_paths('.', '..')
        '..'

        >>> _join_paths('..', '.', '..')
        '../..'

    .. seealso:: _append_step
    """
    nsteps = ft.reduce(_append_step, steps, ())
    if not nsteps:
        return '.'
    else:
        return '/'.join(nsteps)

_forbidden_pstep_attrs = ('get_values', 'Series')
"""
Psteps attributes excluded from magic-creation, because searched by pandas's indexing code.
"""


def _clone_attrs(obj):
    """Clone deeply any collection attributes of the passed-in object."""
    attrs = vars(obj).copy()
    ccsteps = attrs.get('_csteps', None)
    if ccsteps:
        attrs['_csteps'] = ccsteps.copy()
    ctags = attrs['_tags']
    if ctags:
        attrs['_tags'] = attrs['_tags'].copy()
    return attrs


class Pstep(str):

    """
    Automagically-constructed *relocatable* paths for accessing data-tree.

    The "magic" autocreates psteps as they referenced, making writing code
    that access data-tree paths, natural, while at the same time the "model"
    of those tree-data gets discovered.

    Each pstep keeps internally the *name* of a data-tree step, which,
    when created through recursive referencing, concedes with parent's
    branch leading to this step.  That name can be modified with :class:`Pmod`
    so the same data-accessing code can refer to differently-named values
    int the data-tree.

    :ivar dict _csteps:  the child-psteps by their name (default `None`)
    :ivar dict _pmod:    path-modifications used to construct this and
                         relayed to children (default `None`)
    :ivar int _locked:   one of
                         - :const:`Pstep.CAN_RELOCATE` (default),
                         - :const:`Pstep.CAN_RENAME`,
                         - :const:`Pstep.LOCKED` (neither from the above).
    :ivar set _tags:     A set of strings (default `()`)
    :ivar dict _schema:  json-schema data.

    See :meth:`__new__()` for interal constructor.

    **Usage:**

    - Use a :meth:`Pmod.pstep()` to construct a *root* pstep from mappings.
      Specify a string argument to construct a relative pstep-hierarchy.

    - Just referencing (non_private) attributes, creates them.

    - Private attributes and functions (starting with ``_``) exist for
      specific operations (ie for specifying json-schema, or
      for collection all paths).

    - Assignments are only allowed for string-values, or to private attributes::

        >>> p = Pstep()
        >>> p.assignments = 12
        Traceback (most recent call last):
        AssertionError: Cannot assign '12' to '/assignments!

        >>> p._but_hidden = 'Ok'

    - Use :meth:`_paths()` to get all defined paths so far.

    - Construction::

        >>> Pstep()
        ``
        >>> Pstep('a')
        `a`

      Notice that pstesps are surrounded with the back-tick char('`').

    - Paths are created implicitely as they are referenced::

        >>> m = {'a': 1, 'abc': 2, 'cc': 33}
        >>> p = Pstep('a')
        >>> assert m[p] == 1
        >>> assert m[p.abc] == 2
        >>> assert m[p.a321.cc] == 33

        >>> sorted(p._paths())
        ['a/a321/cc', 'a/abc']


    - Any "path-mappings" or "pmods" maybe specified during construction::

        >>> from pandalone.mappings import pmods_from_tuples

        >>> pmods = pmods_from_tuples([
        ...     ('',         'deeper/ROOT'),
        ...     ('/abc',     'ABC'),
        ...     ('/abc/foo', 'BAR'),
        ... ])
        >>> p = pmods.step()
        >>> p.abc.foo
        `BAR`
        >>> p._paths()
        ['deeper/ROOT/ABC/BAR']

    - but exceptions are thrown if mapping any step marked as "locked":

        >>> p.abc.foo._locked  ## 3: CAN_RELOCATE
        3

        >>> p.abc.foo._lock    ## Screams, because `foo` is already mapped.
        Traceback (most recent call last):
        ValueError: Cannot rename/relocate 'foo'-->'BAR' due to LOCKED!

    - .. Warning::
          Creating an empty(`''`) step in some paths will "root" the path::

              >>> p = Pstep()
              >>> _ = p.a1.b
              >>> _ = p.A2
              >>> p._paths()
              ['/A2', '/a1/b']

              >>> _ = p.a1.a2.c
              >>> _ = p.a1.a2 = ''
              >>> p._paths()
              ['/A2', '/a1/b', '/c']

    """

    CAN_RELOCATE = 3
    CAN_RENAME = 1
    LOCKED = 0

    @staticmethod
    def _lockstr(lock):
        if lock >= Pstep.CAN_RELOCATE:
            return 'CAN_RELOCATE'
        if Pstep.LOCKED <= lock < Pstep.CAN_RELOCATE:
            return 'LOCKED'
        return 'LOCKED'

    def __new__(cls, pname='', _proto_or_pmod=None, alias=None):
        """
        Constructs a string with str-content which may comes from the mappings.

        These are the valid argument combinations::

            pname='attr_name',
            pname='attr_name', _alias='Mass [kg]'

            pname='attr_name', _proto_or_pmod=Pmod

            pname='attr_name', _proto_or_pmod=Pstep
            pname='attr_name', _proto_or_pmod=Pstep, _alias='Mass [kg]'


        :param str pname:
                this pstep's name which must coincede with the name of
                the parent-pstep's attribute holding this pstep.
                It is stored at `_orig` and if no `alias` and unmapped by pmod,
                this becomes the `alias`.
        :param Pmod or Pstep _proto_or_pmod:
                It can be either:

                - the mappings for this pstep,
                - another pstep to clone attributes from
                  (used when replacing an existing child-pstep), or
                - None.

                The mappings will apply only if :meth:`Pmod.descend()`
                match `pname` and will derrive the alias.
        :param str alias:
                Will become the super-str object when no mappaings specified
                (`_proto_or_pmod` is a dict from some prototype pstep)
                It gets jsonpointer-escaped if it exists
                (see :func:`pandata.escape_jsonpointer_part()`)
        """
        pmod = None
        attrs = None
        if _proto_or_pmod:
            if isinstance(_proto_or_pmod, Pmod):
                pmod, m_alias = _proto_or_pmod.descend(pname)
                if alias is None and m_alias:
                    # TODO: Add Escape-path TCs.
                    alias = unescape_jsonpointer_part(m_alias)
            else:
                assert isinstance(_proto_or_pmod, Pstep), (
                    'Invalid type(%s) for `_proto_or_pmod`!' % _proto_or_pmod)
                attrs = _clone_attrs(_proto_or_pmod)
        if alias is None:
            alias = pname
        self = str.__new__(cls, alias)
        self.__dict__ = attrs or {
            '_orig': pname,
            '_pmod': pmod,
            '_csteps': None,
            '_locked': Pstep.CAN_RELOCATE,
            '_tags': ()
        }

        return self

    def __make_cstep(self, ckey, alias=None, existing_cstep=None):
        csteps = self._csteps
        if not csteps:
            self._csteps = csteps = {}
        else:
            if existing_cstep is None:
                existing_cstep = csteps.get(ckey, None)
                # Update my mappings for `b` when ``self.b = "foo"``.
                #
                if alias is not None:
                    pmod = self._pmod
                    if pmod:
                        pmod._alias = alias
                    else:
                        self._pmod = Pmod(_alias=alias)
        csteps[ckey] = child = Pstep(ckey, existing_cstep or self._pmod, alias)

        return child

    def __getattr__(self, attr):
        try:
            return str.__getattr__(self, attr)
        except AttributeError as ex:
            if attr.startswith('_') or attr in _forbidden_pstep_attrs:
                raise
            csteps = self._csteps
            child = csteps and csteps.get(attr, None)
            return child or self.__make_cstep(attr)

    def __setattr__(self, attr, value):
        if attr.startswith('_'):
            str.__setattr__(self, attr, value)
        elif isinstance(value, Pstep):
            self.__make_cstep(attr, existing_cstep=value)
        elif isinstance(value, basestring):
            self.__make_cstep(attr, alias=value)
        else:
            raise self._ex_invalid_assignment(attr, value)

    def __dir__(self):
        d = super(str, self).__dir__()
        if self._csteps:
            d = sorted(d + list(self._csteps.keys()))
        return d

    def _ex_invalid_assignment(self, cpname, value):
        msg = "Cannot assign '%s' to '%s/%s!"
        return AssertionError(msg % (value, self, cpname))

    def __repr__(self):
        return '`%s`' % self

    @property
    def _locked(self):
        """
        Gets `_locked` internal flag or scream on set, when step already renamed/relocated

        Prefer using one of :attr:`_fix` or :attr:`_lock` instead.

        :param locked:  One of :attr:`CAN_RELOCATE`, :attr:`CAN_RENAME`,
                        :attr:`LOCKED`.
        :raise: ValueError when stricter lock-value on a renamed/relocated pstep
        """
        return vars(self)['_locked']

    @_locked.setter
    def _locked(self, lock_state):
        if self != self._orig:
            if lock_state < Pstep.CAN_RENAME or (
                    lock_state < Pstep.CAN_RELOCATE and '/' in self):
                msg = "Cannot rename/relocate '%s'-->'%s' due to %s!"
                raise ValueError(
                    msg % (self._orig, self, Pstep._lockstr(lock_state)))
        vars(self)['_locked'] = int(lock_state)

    @property
    def _fix(self):
        """Sets :attr:`locked` = `CAN_RENAME`.
        :return: self
        :raise: ValueError if step has been relocated pstep
        """
        self._locked = Pstep.CAN_RENAME
        return self

    @property
    def _lock(self):
        """Set :attr:`locked` = `LOCKED`.
        :return: self, for chained use
        :raise: ValueError if step has been renamed/relocated pstep
        """
        self._locked = Pstep.LOCKED
        return self

    def _tag(self, tag):
        """Add a "tag" for this pstep.

        :return: self, for chained use
        """
        tags = self._tags
        if tags:
            tags.add(tag)
        else:
            self._tags = set([tag])

        return self

    def _tag_remove(self, tag):
        """Delete a "tag" from this pstep.

        :return: self, for chained use
        """
        tags = self._tags
        if tags:
            tags.discard(tag)

        return self

    def _steps(self, keys=False, tag=None):
        csteps = self._csteps
        if not csteps:
            return []
        if keys:
            return csteps.keys()
        return csteps.values()

    def _paths(self, with_orig=False, tag=None):
        """
        Return all children-paths (str-list) constructed so far, in a list.

        :param bool with_orig: wheter to include also orig-path, for debug.
        :param str tag:        If not 'None', fetches all paths with `tag`
                               in their last step.
        :rtype: [str]


        Examples::

              >>> p = Pstep()
              >>> _ = p.a1._tag('inp').b._tag('inp').c
              >>> _ = p.a2.b2

              >>> p._paths()
              ['/a1/b/c', '/a2/b2']

              >>> p._paths(tag='inp')
              ['/a1', '/a1/b']


        For debugging set `with_orig` to `True`::

            >>> pmods = pmods_from_tuples([
            ...     ('',         'ROOT'),
            ...     ('/a',     'A/AA'),
            ... ])
            >>> p = pmods.step()
            >>> _ = p.a.b
            >>> p._paths(with_orig=True)
             ['(-->ROOT)/(a-->A/AA)/b']

        """
        def append_orig(s):
            orig = s._orig
            return '(%s-->%s)' % (orig, s) if s != orig else s

        paths = []
        for path in self._iter_hierarchy():
            send = path[-1]
            if (tag and tag in send._tags) or (not tag and not send._csteps):
                if with_orig:
                    path = [append_orig(s) for s in path]
                paths.append(_join_paths(*path))

        return sorted(set(paths))

    def _derrive_map_tuples(self):
        """
        Recursively extract ``(cmap --> alias)`` pairs from the pstep-hierarchy.

        :param list pairs:             Where to append subtree-paths built.
        :param tuple prefix_steps:     branch currently visiting
        :rtype: [(str, str)]
        """
        def orig_paths(psteps):
            return [p._orig for p in psteps]

        map_pairs = ((_join_paths(*orig_paths(p)), str(p[-1]))
                     for p in self._iter_hierarchy())
        return sorted(map_pairs, key=lambda p: p[0])

    def _iter_hierarchy(self, prefix_steps=()):
        """
        Breadth-first traversing of pstep-hierarchy.

        :param tuple prefix_steps:     Builds here branch currently visiting.
        :return: yields the visited pstep along with its path (including it)
        :rtype: (Pstep, [Pstep])
        """
        prefix_steps += (self, )
        yield prefix_steps

        csteps = self._csteps
        if csteps:
            for v in csteps.values():
                for p in v._iter_hierarchy(prefix_steps):
                    yield p

    @property
    def _schema(self):
        """Updates json-schema-v4 on this pstep (see :class:`JSchema`)."""

        # Lazy create it
        #    (clients should check before`_schema_exists()`)
        #
        sdict = vars(self)
        jschema = sdict.get('_schema')
        if jschema is None:
            sdict['_schema'] = jschema = JSchema()

        return jschema

    def _schema_exists(self):
        """Always use this to avoid needless schema-instantiations."""
        return '_schema' in vars(self)


def pstep_from_df(columns_df, name_col='names'):
    """
    Creates a :class:`Pstep` instances from a dataframe.

    :param pd.DataFrame columns_df:
            pstep's mapped-names in `name_col` column, indexed by paths, and
            any additional pstep-attributes in the rest columns.

            example::

                ========  =========  ===================
                paths     names      renames
                ========  =========  ===================
                /A        foo        ['FOO', 'LL']
                /B        bar        []
                ========  =========  ===================
    """
    p = pmods_from_tuples(zip(columns_df.index, columns_df[name_col])).step()
    cdf = columns_df.drop(name_col, axis=1, errors='ignore')
    attributes = cdf.columns
    for rows in cdf.itertuples():
        path, attr_values = rows[1], rows[1:]
        cstep = getattr(p, path[1:])
        for attr, aval in zip(attributes, attr_values):
            setattr(cstep, '_%s' % attr, aval)

    return p


if __name__ == '__main__':  # pragma: no cover
    raise NotImplementedError



#! python
# -*- coding: UTF-8 -*-
#
# Copyright 2015 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
from __future__ import division, unicode_literals

import doctest
import logging
import os
from pandalone.utils import fullmatch_py2, LoggerWriter
import re
import sys
import unittest

import ddt

import pandalone.utils as utils


@unittest.skipIf(sys.version_info < (3, 4), "Doctests are made for py >= 3.3")
class Doctest(unittest.TestCase):

    def test_doctests(self):
        failure_count, test_count = doctest.testmod(
            utils, optionflags=doctest.NORMALIZE_WHITESPACE)
        self.assertGreater(test_count, 0, (failure_count, test_count))
        self.assertEquals(failure_count, 0, (failure_count, test_count))


class TestUtils(unittest.TestCase):

    def _assert_expansion(self, regex, s, template):
        m1 = re.fullmatch(regex, s)
        s1 = m1.expand(template) if m1 else '<NO-MATCH>'
        m2 = fullmatch_py2(regex, s)
        s2 = m2.expand(template) if m2 else '<NO-MATCH>'
        self.assertEqual(s1, s2, "\n  PY-3: '%s' \n  PY-2: '%s' " % (s1, s2))

    @unittest.skipIf(not 'fullmatch' in dir(re), "Platform has no 'fullmatch() to compare with.")
    def test_fullmatch(self):
        self._assert_expansion('.*', 'foo', r'A')
        self._assert_expansion('(.*)', 'foo', r'A_\1')
        self._assert_expansion('(.*)', 'foo', r'A_\g<0>')

        self._assert_expansion('a.*', 'afoo&', r'A')
        self._assert_expansion('a(\w*)', 'afoo&', r'A_\1')
        self._assert_expansion('a(\w*)', 'afoo&', r'A_\g<0>')

    # Not fixed for performance,
    # for full-solution, see:
    #   http://stackoverflow.com/questions/30212413/backport-python-3-4s-regular-expression-fullmatch-to-python-2")
    @unittest.expectedFailure
    def test_fullmatch_hard(self):
        self._assert_expansion(".*?", "Hello", '\g<0>')

    def test_make_unique_filename(self):
        fname = '/dir/_NOT_EXISTS_'
        self.assertEqual(utils.make_unique_filename(fname), fname)

        os.chdir(os.path.dirname(__file__))
        fname = os.path.basename(__file__)
        bname, e = os.path.splitext(fname)
        fname1 = '%s1%s' % (bname, e)
        self.assertEqual(utils.make_unique_filename(fname), fname1)

        fname = os.path.join('..', 'tests', os.path.basename(__file__))
        fname2 = os.path.join('..', 'tests', fname1)
        self.assertEqual(utils.make_unique_filename(fname), fname2)

    @unittest.skipIf(sys.version_info < (3, 4), "TC.assertLog() not exists!")
    def test_LogWriter_smoke(self):
        logname = "foobar"
        log = logging.getLogger(logname)
        level = logging.INFO
        lw = LoggerWriter(log, level)
        with self.assertLogs(logname, level):
            lw.write('Hehe')
        lw.flush()


@unittest.skipIf(os.name != 'nt', "Cannot test Windows paths.")
@ddt.ddt
class TPath2Url(unittest.TestCase):

    @ddt.data(
        ('foo', 'file:%s/foo'),
        ('foo/', 'file:%s/foo/'),
        ('foo/bar', 'file:%s/foo/bar'),

        ('foo\\', 'file:%s/foo/'),
        ('foo\\bar', 'file:%s/foo/bar'),

        ('./foo', 'file:%s/foo'),
        ('./foo/', 'file:%s/foo/'),
        ('./foo/bar', 'file:%s/foo/bar'),

        ('.\\foo', 'file:%s/foo'),
        ('.\\foo\\', 'file:%s/foo/'),
        ('.\\foo/bar', 'file:%s/foo/bar'),
    )
    def test_relative(self, case):
        from future.moves.urllib.request import pathname2url
        path, url = case
        cwd = pathname2url(os.getcwd())
        self.assertEqual(utils.path2url(path), url % cwd, path)

    @ddt.data(
        ('/', 'file:///'),
        ('/foo', 'file:///foo'),
        ('/foo/', 'file:///foo/'),
        ('/foo/bar', 'file:///foo/bar'),

        ('\\', 'file:///'),
        ('\\foo', 'file:///foo'),
        ('\\foo\\', 'file:///foo/'),
        ('\\foo\\bar', 'file:///foo/bar'),
    )
    def test_absolute(self, case):
        path, url = case
        self.assertEqual(utils.path2url(path), url, path)

    @ddt.data(
        ('D:\\', 'file:///D:/'),
        ('d:\\foo', 'file:///d:/foo'),
        ('C:\\foo\\', 'file:///C:/foo/'),
        ('c:\\foo\\bar', 'file:///c:/foo/bar'),

        ('D:', 'file:///D:/'),  # NOTE: destroy Windows per drive cwd!
        ('d:foo', 'file:///d:/foo'),
        ('C:foo\\', 'file:///C:/foo/'),
        ('c:foo\\bar', 'file:///c:/foo/bar'),
    )
    def test_drive(self, case):
        path, url = case
        self.assertEqual(utils.path2url(path), url, path)

    @ddt.data(
        ('d:../foo', 'file:///foo'),
        ('D:../../foo', 'file:///foo'),
    )
    def test_corner_cases(self, case):
        path, url = case
        self.assertEqual(utils.path2url(path), url, path)

    @ddt.data(
        ('/%REL%', 'file:///RRR'),
        ('/$REL', 'file:///RRR'),
        ('/${REL}', 'file:///RRR'),

        ('/%REL%/', 'file:///RRR/'),
        ('/$REL/', 'file:///RRR/'),
        ('/${REL}/', 'file:///RRR/'),

        ('%ABS%', 'file:///AAA'),
        ('$ABS', 'file:///AAA'),
        ('${ABS}', 'file:///AAA'),

        ('%ABS%/', 'file:///AAA/'),
        ('$ABS/', 'file:///AAA/'),
        ('${ABS}/', 'file:///AAA/'),
    )
    def test_expansion_vars(self, case):
        path, url = case
        os.environ['REL'] = 'RRR'
        os.environ['ABS'] = '/AAA'
        self.assertEqual(utils.path2url(path, 1, 1), url, path)

    @ddt.data(
        ('~/a', '/a'),
        ('~\\a', '/a'),
        ('~/a/', '/a/'),
        ('~\\a\\', '/a/'),
    )
    def test_expansion_user(self, case):
        path, end = case
        url = utils.path2url(path, 0, 1)
        self.assertFalse(url.startswith('~'), (url, path))
        self.assertTrue(url.endswith(end), (url, path))

        url = utils.path2url(path, 0, 0)
        # NO, turned into absolute!
        ##self.assertTrue(url.startswith('~'), (url, path))

    @ddt.data(
        ('http://ser/D:\\', 'http://ser/D:/'),
        ('http://ser/d:\\foo', 'http://ser/d:/foo'),
        ('http://ser/foo/', 'http://ser/foo/'),
    )
    def test_remote(self, case):
        path, url = case
        self.assertEqual(utils.path2url(path), url, path)



#!/usr/bin/env python
#
# Copyright 2014-2015 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""Authoritative project's version-info"""
__version__ = version = "0.2.2"  # PEP 440 versioning. Also update README.rst's.
__updated__ = "2017-02-07 04:05:41"



#! python
# -*- coding: UTF-8 -*-
#
# Copyright 2014-2015 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
import os
from pandalone import xlsutils
from tests._tutils import (
    init_logging, TemporaryDirectory, xw_no_save_Workbook,
    xw_close_workbook)
import unittest

import pandas as pd


log = init_logging(__name__)
is_excel_installed = xlsutils.check_excell_installed()


def from_my_path(*parts):
    return os.path.join(os.path.dirname(__file__), *parts)


@unittest.skipIf(not is_excel_installed, "Cannot test xlwings without MS Excel.")
class TestExcel(unittest.TestCase):

    def test_build_excel(self):
        with TemporaryDirectory() as tmpdir:
            wb_inp_fname = from_my_path('empty.xlsx')
            wb_out_fname = from_my_path(tmpdir, 'ExcelRunner.xlsm')
            vba_wildcard = from_my_path('..', 'pandalone', 'excel', '*.vba')
            try:
                wb = xlsutils.import_vba_into_excel_workbook(
                    vba_wildcard, wb_inp_fname, wb_out_fname)
            finally:
                if 'wb' in locals():
                    xw_close_workbook(wb)

    def test_xlwings_smoketest(self):
        import xlwings as xw
        sheetname = 'shitt'
        addr = 'f6'
        table = pd.DataFrame([[1, 2], [True, 'off']], columns=list('ab'))
        with xw_no_save_Workbook() as wb:
            xw.Sheet(1).name = sheetname
            xw.Range(addr).value = table


if __name__ == "__main__":
    unittest.main()



#! python
# -*- coding: UTF-8 -*-
#
# Copyright 2014-2015 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

from __future__ import division, print_function, unicode_literals

import doctest
import json
import sys
import unittest

from jsonschema.exceptions import RefResolutionError

import numpy as np
import numpy.testing as npt
from pandalone import pandata
from pandalone.pandata import (JSONCodec, iter_jsonpointer_parts,
                               iter_jsonpointer_parts_relaxed, escape_jsonpointer_part,
                               unescape_jsonpointer_part)
import pandas as pd
from tests._tutils import init_logging


log = init_logging(__name__)


@unittest.skipIf(sys.version_info < (3, 4), "Doctests are made for py >= 3.3")
class Doctest(unittest.TestCase):

    def test_doctests(self):
        failure_count, test_count = doctest.testmod(
            pandata,
            optionflags=doctest.NORMALIZE_WHITESPACE)  # | doctest.ELLIPSIS)
        self.assertGreater(test_count, 0, (failure_count, test_count))
        self.assertEquals(failure_count, 0, (failure_count, test_count))


class TestJSONCodec(unittest.TestCase):

    def test_lists(self):
        o = [1, 2, '3', 'fdgdg', None, []]
        s = json.dumps(o, cls=JSONCodec.Encoder)
        oo = json.loads(s, cls=pandata.JSONCodec.Decoder)
        self.assertEqual(oo, o)

    def test_dict(self):
        o = {'a': 1, 2: 'bb', 3: [], 4: None}
        s = json.dumps(o, cls=JSONCodec.Encoder)
        oo = json.loads(s, cls=pandata.JSONCodec.Decoder)
        self.assertEqual(oo, o)

    def test_Numpy(self):
        o = np.random.randn(6, 2)
        s = json.dumps(o, cls=JSONCodec.Encoder)
        oo = json.loads(s, cls=pandata.JSONCodec.Decoder)
        npt.assert_array_equal(oo, o)

    def test_DataFrame(self):
        o = pd.DataFrame(np.random.randn(10, 2))
        s = json.dumps(o, cls=JSONCodec.Encoder)
        oo = json.loads(s, cls=pandata.JSONCodec.Decoder)
        npt.assert_array_equal(oo, o)

    def test_Series(self):
        o = pd.Series({'a': 1, 2: 22})
        s = json.dumps(o, cls=JSONCodec.Encoder)
        oo = json.loads(s, cls=pandata.JSONCodec.Decoder)
        npt.assert_array_equal(oo, o)

    @unittest.skip(('TODO: Cannot test for recursive-equality with pandas.'))
    def test_mix(self):
        obj_list = [3.14,
                    {
                        'aa': pd.DataFrame([]),
                        2: np.array([]),
                        33: {'foo': 'bar'},
                    },
                    pd.DataFrame(np.random.randn(10, 2)),
                    ('b', pd.Series({})),
                    ]
        for o in obj_list + [obj_list]:
            s = json.dumps(o, cls=JSONCodec.Encoder)
            oo = json.loads(s, cls=pandata.JSONCodec.Decoder)
            self.assertTrue(pandata.trees_equal(o, oo), (o, oo))


class TestJsonPointer(unittest.TestCase):

    def test_jsonpointer_escape_parts(self):
        def un_esc(part):
            return unescape_jsonpointer_part(escape_jsonpointer_part(part))
        part = 'hi/there'
        self.assertEqual(un_esc(part), part)
        part = 'hi~there'
        self.assertEqual(un_esc(part), part)
        part = '/hi~there/'
        self.assertEqual(un_esc(part), part)

    def test_iter_jsonpointer_empty(self):
        self.assertListEqual(list(iter_jsonpointer_parts('')), [])
        self.assertListEqual(list(iter_jsonpointer_parts_relaxed('')), [''])

    def test_iter_jsonpointer_root(self):
        self.assertListEqual(list(iter_jsonpointer_parts('/')), [''])
        self.assertListEqual(
            list(iter_jsonpointer_parts_relaxed('/')), ['', ''])

    def test_iter_jsonpointer_regular(self):
        self.assertListEqual(list(iter_jsonpointer_parts('/a')), ['a'])
        self.assertListEqual(
            list(iter_jsonpointer_parts_relaxed('/a')), ['', 'a'])

        self.assertListEqual(list(iter_jsonpointer_parts('/a/b')), ['a', 'b'])
        self.assertListEqual(
            list(iter_jsonpointer_parts_relaxed('/a/b')), ['', 'a', 'b'])

    def test_iter_jsonpointer_folder(self):
        self.assertListEqual(list(iter_jsonpointer_parts('/a/')), ['a', ''])
        self.assertListEqual(
            list(iter_jsonpointer_parts_relaxed('/a/')), ['', 'a', ''])

    def test_iter_jsonpointer_non_absolute(self):
        with self.assertRaises(RefResolutionError):
            list(iter_jsonpointer_parts('a'))
        with self.assertRaises(RefResolutionError):
            list(iter_jsonpointer_parts('a/b'))

    def test_iter_jsonpointer_None(self):
        with self.assertRaises(AttributeError):
            list(iter_jsonpointer_parts(None))
        with self.assertRaises(AttributeError):
            list(iter_jsonpointer_parts_relaxed(None))

    def test_iter_jsonpointer_with_spaces(self):
        self.assertListEqual(
            list(iter_jsonpointer_parts('/ some ')), [' some '])
        self.assertListEqual(
            list(iter_jsonpointer_parts('/ some /  ')), [' some ', '  '])

        self.assertListEqual(
            list(iter_jsonpointer_parts_relaxed(' some ')), [' some '])
        self.assertListEqual(
            list(iter_jsonpointer_parts_relaxed(' some /  ')), [' some ', '  '])

    def test_iter_jsonpointer_massive(self):
        cases = [
            ('/a', ['a']),
            ('/a/', ['a', '']),
            ('/a/b', ['a', 'b']),
            ('/a/b/', ['a', 'b', '']),
            ('/a//b', ['a', '', 'b']),
            ('/a/../b', ['a', '..', 'b']),
            ('/', ['']),
            ('', []),
            ('/ some ', [' some ']),
            ('/ some /', [' some ', '']),
            ('/ some /  ', [' some ', '  ']),
            ('/ some /  /', [' some ', '  ', '']),
            (None, AttributeError),
            ('a', RefResolutionError),
        ]
        for i, (inp, out) in enumerate(cases):
            msg = 'case #%i' % i
            try:
                if issubclass(out, Exception):
                    with self.assertRaises(out, msg=msg):
                        list(iter_jsonpointer_parts(inp))
                continue
            except TypeError as ex:
                if ex.args[0].startswith('issubclass()'):
                    self.assertEqual(
                        list(iter_jsonpointer_parts(inp)), out, msg)
                else:
                    raise ex

    def test_iter_jsonpointer_relaxed_massive(self):
        cases = [
            ('/a', ['', 'a']),
            ('/a/', ['', 'a', '']),
            ('/a/b', ['', 'a', 'b']),
            ('/a/b/', ['', 'a', 'b', '']),
            ('/a//b', ['', 'a', '', 'b']),
            ('/', ['', '']),
            ('', ['']),
            ('/ some ', ['', ' some ']),
            ('/ some /', ['', ' some ', '']),
            ('/ some /  ', ['', ' some ', '  ']),
            (None, AttributeError),
            ('a', ['a']),
            ('a/', ['a', '']),
            ('a/b', ['a', 'b']),
            ('a/b/', ['a', 'b', '']),
            ('a/../b/.', ['a', '..', 'b', '.']),
            ('a/../b/.', ['a', '..', 'b', '.']),
            (' some ', [' some ']),
            (' some /', [' some ', '']),
            (' some /  ', [' some ', '  ']),
            (' some /  /', [' some ', '  ', '']),
        ]
        for i, (inp, out) in enumerate(cases):
            msg = 'case #%i' % i
            try:
                if issubclass(out, Exception):
                    with self.assertRaises(out, msg=msg):
                        list(iter_jsonpointer_parts_relaxed(inp))
                continue
            except TypeError as ex:
                if ex.args[0].startswith('issubclass()'):
                    self.assertEqual(
                        list(iter_jsonpointer_parts_relaxed(inp)), out, msg)
                else:
                    raise ex

    def test_resolve_jsonpointer_existing(self):
        doc = {
            'foo': 1,
            'bar': [11, {'a': 222}]
        }

        path = '/foo'
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), 1)

        path = '/bar/0'
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), 11)

        path = '/bar/1/a'
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), 222)

    def test_resolve_jsonpointer_sequence(self):
        doc = [1, [22, 33]]

        path = '/0'
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), 1)

        path = '/1'
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), [22, 33])

        path = '/1/0'
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), 22)
        path = '/1/1'
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), 33)

    def test_resolve_jsonpointer_missing_screams(self):
        doc = {}

        path = '/foo'
        with self.assertRaises(RefResolutionError):
            pandata.resolve_jsonpointer(doc, path)

    def test_resolve_jsonpointer_empty_path(self):
        doc = {}
        path = ''
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), doc)

        doc = {'foo': 1}
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), doc)

    def test_resolve_jsonpointer_examples_from_spec(self):
        def _doc():
            return {
                r"foo": ["bar", r"baz"],
                r"": 0,
                r"a/b": 1,
                r"c%d": 2,
                r"e^f": 3,
                r"g|h": 4,
                r"i\\j": 5,
                r"k\"l": 6,
                r" ": 7,
                r"m~n": 8
            }
        cases = [
            (r"",            _doc()),
            (r"/foo",        ["bar", "baz"]),
            (r"/foo/0",      "bar"),
            (r"/",           0),
            (r"/a~1b",       1),
            (r"/c%d",      2),
            (r"/e^f",      3),
            (r"/g|h",      4),
            (r"/i\\j",      5),
            (r"/k\"l",      6),
            (r"/ ",        7),
            (r"/m~0n",       8),
        ]
        for path, exp in cases:
            doc = _doc()
            self.assertEqual(pandata.resolve_jsonpointer(doc, path), exp)

    def test_set_jsonpointer_empty_doc(self):
        doc = {}
        path = '/foo'
        value = 'value'
        pandata.set_jsonpointer(doc, path, value)
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), value)

        doc = {}
        path = '/foo/bar'
        value = 'value'
        pandata.set_jsonpointer(doc, path, value)
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), value)

    def test_resolve_jsonpointer_root_path_only(self):
        doc = {}
        path = '/'
        with self.assertRaises(RefResolutionError):
            self.assertEqual(pandata.resolve_jsonpointer(doc, path), doc)

        doc = {'foo': 1}
        with self.assertRaises(RefResolutionError):
            self.assertEqual(pandata.resolve_jsonpointer(doc, path), doc)

        doc = {'': 1}
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), doc[''])

    def test_set_jsonpointer_replace_value(self):
        doc = {'foo': 'bar', 1: 2}
        path = '/foo'
        value = 'value'
        pandata.set_jsonpointer(doc, path, value)
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), value)
        self.assertEqual(doc[1], 2)

        doc = {'foo': 1, 1: 2}
        path = '/foo'
        value = 'value'
        pandata.set_jsonpointer(doc, path, value)
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), value)
        self.assertEqual(doc[1], 2)

        doc = {'foo': {'bar': 1}, 1: 2}
        path = '/foo'
        value = 2
        pandata.set_jsonpointer(doc, path, value)
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), value)
        self.assertEqual(doc[1], 2)

    def test_set_jsonpointer_append_path(self):
        doc = {'foo': 'bar', 1: 2}
        path = '/foo/bar'
        value = 'value'
        pandata.set_jsonpointer(doc, path, value)
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), value)
        self.assertEqual(doc[1], 2)

        doc = {'foo': 'bar', 1: 2}
        path = '/foo/bar/some/other'
        value = 'value'
        pandata.set_jsonpointer(doc, path, value)
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), value)
        self.assertEqual(doc[1], 2)

    def test_set_jsonpointer_append_path_preserves_intermediate(self):
        doc = {'foo': {'bar': 1}, 1: 2}
        path = '/foo/foo2'
        value = 'value'
        pandata.set_jsonpointer(doc, path, value)
        print(doc)
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), value)
        self.assertEqual(doc[1], 2)
        self.assertEqual(pandata.resolve_jsonpointer(doc, '/foo/bar'), 1)

    def test_set_jsonpointer_missing(self):
        doc = {'foo': 1, 1: 2}
        path = '/foo/bar'
        value = 'value'
        pandata.set_jsonpointer(doc, path, value)
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), value)
        self.assertEqual(doc[1], 2)

        doc = {'foo': 1, 1: 2}
        path = '/foo/bar/some/other'
        value = 'value'
        pandata.set_jsonpointer(doc, path, value)
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), value)
        self.assertEqual(doc[1], 2)

    def test_set_jsonpointer_sequence(self):
        doc = [1, 2]
        path = '/1'
        value = 'value'
        pandata.set_jsonpointer(doc, path, value)
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), value)

        doc = [1, 2]
        path = '/1/foo/bar'
        value = 'value'
        pandata.set_jsonpointer(doc, path, value)
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), value)

    def test_set_jsonpointer_sequence_insert_end(self):
        doc = [0, 1]
        path = '/2'
        value = 'value'
        pandata.set_jsonpointer(doc, path, value)
        self.assertEqual(pandata.resolve_jsonpointer(doc, path), value)
        self.assertEqual(pandata.resolve_jsonpointer(doc, '/0'), 0)
        self.assertEqual(pandata.resolve_jsonpointer(doc, '/1'), 1)

        doc = [0, 1]
        path = '/-'
        value = 'value'
        pandata.set_jsonpointer(doc, path, value)
        self.assertEqual(pandata.resolve_jsonpointer(doc, '/2'), value)
        self.assertEqual(pandata.resolve_jsonpointer(doc, '/0'), 0)
        self.assertEqual(pandata.resolve_jsonpointer(doc, '/1'), 1)

    def test_set_jsonpointer_sequence_out_of_bounds(self):
        doc = [0, 1]
        path = '/3'
        value = 'value'
        with self.assertRaises(RefResolutionError):
            pandata.set_jsonpointer(doc, path, value)

    def test_set_jsonpointer_sequence_with_str_screams(self):
        doc = [0, 1]
        path = '/str'
        value = 'value'
        with self.assertRaises(RefResolutionError):
            pandata.set_jsonpointer(doc, path, value)

    def test_build_all_jsonpaths(self):
        from jsonschema._utils import load_schema
        schema = dict(properties=dict(a=dict(properties=dict(b={}))))
        paths = pandata.build_all_jsonpaths(schema)
        print('\n'.join(paths))
        # TODO: build and check_all_paths support $ref
        self.assertIn('/a/b', paths)


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



#! python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2015 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

from __future__ import division, unicode_literals

import unittest

from jsonschema.exceptions import ValidationError

import numpy.testing as npt
from pandalone import pandata
from pandalone.pandata import PandelVisitor
import pandas as pd

from ._tutils import assertRaisesRegex


class TestJsonPath(unittest.TestCase):

    def test_ModelMaker_merge(self):
        class MyMaker(pandata.Pandel):

            def _get_json_schema(self, is_prevalidation):
                return {
                    '$schema': 'http://json-schema.org/draft-04/schema#',
                    'type': ['object'],
                    'required': [] if is_prevalidation else ['a', 'b'],
                    'properties': {
                        'a': {'type': ['number', 'array']},
                        'b': {'type': ['number', 'array']},
                        'c': {'type': ['number', 'array']},
                    }
                }

        # Maps
        mm = MyMaker()

        (s1, s2) = {'a': 1}, {'a': 2}
        mdl = mm._clone_and_merge_submodels(s1, s2)
        self.assertEqual(mdl['a'], 2)

        (s1, s2) = {'a': 1, 'b': 2}, {'a': 11, 'c': 3}
        mdl = mm._clone_and_merge_submodels(s1, s2)
        self.assertEqual(mdl['a'], 11)
        self.assertEqual(mdl['b'], 2)
        self.assertEqual(mdl['c'], 3)

        # DataFranes
        (s1, s2) = pd.DataFrame({'a': [1, 2], 'b': [3, 4]}), {
            'a': [11, 22], 'c': [5, 6]}
        mdl = mm._clone_and_merge_submodels(s1, s2)
        npt.assert_array_equal(mdl['a'], [11, 22])
        npt.assert_array_equal(mdl['b'], [3, 4])
        npt.assert_array_equal(mdl['c'], [5, 6])

        # Series
        (s1, s2) = pd.Series({'a': 1, 'b': 3}), {'a': 11, 'c': 5}
        mdl = mm._clone_and_merge_submodels(s1, s2)
        self.assertEqual(mdl['a'], 11)
        self.assertEqual(mdl['b'], 3)
        self.assertEqual(mdl['c'], 5)

        # Sequences
        l2 = [4, 5]
        (s1, s2) = [1, 2, 3], l2
        mdl = mm._clone_and_merge_submodels(s1, s2)
        npt.assert_array_equal(mdl, l2)

        # Sequences-->Map
        l2 = [4, 5, {'a': 11, 'c': 33}]
        (s1, s2) = [1, 2, {'a': 1, 'b': 2}], l2
        mdl = mm._clone_and_merge_submodels(s1, s2)
        npt.assert_array_equal(mdl, l2)
        obj = mdl[2]
        self.assertEqual(obj['a'], 11)
        self.assertEqual(obj['c'], 33)
        self.assertTrue('b' not in obj)

        # Map-->Sequence
        l2 = [4, 5, {'a': 11, 'c': 3}]
        (s1, s2) = {'a': 1, 'b': [1, 2]}, {'b': l2, 'c': 3}
        mdl = mm._clone_and_merge_submodels(s1, s2)
        self.assertEqual(mdl['a'], 1)
        self.assertEqual(mdl['c'], 3)
        npt.assert_array_equal(mdl['b'], l2)

        # Map-->Map
        (s1, s2) = {'a': 1, 'b': {'aa': -1, 'bb': -2},
                    'c': 3}, {'a': 11, 'b': {'aa': -11, 'cc': -33}, 'd': 44}
        mdl = mm._clone_and_merge_submodels(s1, s2)
        self.assertEqual(mdl['a'], 11)
        self.assertEqual(mdl['c'], 3)
        self.assertEqual(mdl['d'], 44)
        self.assertEqual(mdl['b'], {'aa': -11, 'bb': -2, 'cc': -33})

    def test_ModelMaker_build(self):
        class MyMaker(pandata.Pandel):

            def _get_json_schema(self, is_prevalidation):
                return {
                    '$schema': 'http://json-schema.org/draft-04/schema#',
                    'required': [] if is_prevalidation else ['a', 'b'],
                    'properties': {
                        'a': {'type': 'string'},
                        'b': {'type': 'number'},
                        'c': {'type': 'number'},
                    }
                }

        mm = MyMaker()
        mm.add_submodel({'a': 'foo', 'b': 1})
        mm.add_submodel({'a': 'bar', 'c': 2})

        mdl = mm.build()

        self.assertEqual(mdl['a'], 'bar')
        self.assertEqual(mdl['b'], 1)
        self.assertEqual(mdl['c'], 2)

        mm = MyMaker()  # Invalid submodel['b'], must be a number
        mm.add_submodel({'a': 'foo', 'b': 'string'})
        assertRaisesRegex(
            self, ValidationError, "Failed validating u?'type' in schema\[u?'properties']\[u?'b']", mm.build)

    def test_validate_object_or_pandas(self):
        schema = {
            'type': ['object'],
        }
        pv = PandelVisitor(schema)

        pv.validate({'foo': 'bar'})
        pv.validate(pd.Series({'foo': 'bar', 'foofoo': 'bar'}))
        pv.validate(pd.DataFrame({'foo': [1, 2], 'foofoo': [3, 4]}))

        with assertRaisesRegex(self, ValidationError, "\[1, 2, 3\] is not of type u?'object'"):
            pv.validate([1, 2, 3])

    def test_rule_requiredProperties_rule_for_pandas(self):
        schema = {
            'type': ['object'],
            'required': ['foo']
        }
        pv = PandelVisitor(schema)

        pv.validate({'foo': 'bar'})
        with assertRaisesRegex(self, ValidationError, "'foo' is a required property"):
            pv.validate({'foofoo': 'bar'})

        pv.validate(pd.Series({'foo': 'bar', 'foofoo': 'bar'}))
        with assertRaisesRegex(self, ValidationError, "'foo' is a required property"):
            pv.validate(pd.Series({'foofoo': 'bar'}))

        pv.validate(pd.DataFrame({'foo': [1, 2], 'foofoo': [3, 4]}))
        with assertRaisesRegex(self, ValidationError, "'foo' is a required property"):
            pv.validate(pd.DataFrame({'foofoo': [1, 2], 'bar': [3, 4]}))

    def test_rule_additionalProperties_for_pandas(self):
        schema = {
            'type': ['object'],
            'additionalProperties': False,
            'properties': {
                'foo': {},
            }
        }
        pv = PandelVisitor(schema)

        pv.validate({'foo': 1})
        with assertRaisesRegex(self, ValidationError, "Additional properties are not allowed \(u?'bar' was unexpected\)"):
            pv.validate({'foo': 1, 'bar': 2})

        pv.validate(pd.Series({'foo': 1}))
        with assertRaisesRegex(self, ValidationError, "Additional properties are not allowed \(u?'bar' was unexpected\)"):
            pv.validate(pd.Series({'foo': 1, 'bar': 2}))

        pv.validate(pd.DataFrame({'foo': [1]}))
        with assertRaisesRegex(self, ValidationError, "Additional properties are not allowed \(u?'bar' was unexpected\)"):
            pv.validate(pd.DataFrame({'foo': [1], 'bar': [2]}))


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# Pandalone tasks.
import pandas as pd


def task_read_input():
    def read_input_csv(dependencies):
        df = pd.read_csv(next(iter(dependencies)))
        return df
    
    return {
        'actions': [read_input_csv],
        'file_dep': ['input.csv'],
    }
